//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				: a lot for DX8
//-----------------------------------------------------------------------------

#ifndef _G3DD_H_
#define _G3DD_H_


#include <windows.h>
#include <windowsx.h>
#include <basetsd.h>
#include <mmsystem.h>
#include <stdio.h>
#include <tchar.h>
#include <atlbase.h>


#include <D3D8.h>
#include <D3DX8.h>
#include <DXERR8.h>
#include <D3D8TYPES.h>
#include <dshow.h>
//#include <g3def.h>
#include <g3rgb.h>
#include <g3mat.h>
#include <g3font.h>

#include <g3utils.h>

#include <g3gvtx.h>
#include <g3trace.h>
//#include <g3font.h>

// defines for FILTER STATE
#define DD_MIPMAP_NONE			D3DTEXF_NONE 
#define DD_MIPMAP_POINT			D3DTEXF_POINT
#define DD_MIPMAP_LINEAR		D3DTEXF_LINEAR
#define DD_FILTER_POINT			D3DTEXF_POINT
#define DD_FILTER_LINEAR		D3DTEXF_LINEAR
#define DD_FILTER_ANISOTROPIC	D3DTEXF_ANISOTROPIC
// enumeration max
#define DD_DRIVERMAX			4
#define DD_DEVICEMAX			8
#define DD_MODEMAX				128
// cache index format
#define DD_CACHEIDXBIT		(10)
#define DD_CACHEIDXMASK		((1<<DD_CACHEIDXBIT)-1)
#define DD_CACHEIDXMASK1	(DD_CACHEIDXMASK)
#define DD_CACHEIDXMASK2	(DD_CACHEIDXMASK<<(DD_CACHEIDXBIT))
#define DD_CACHEIDXMASK3	(DD_CACHEIDXMASK<<(DD_CACHEIDXBIT+DD_CACHEIDXBIT))
#define DD_GETCACHEIDX1(b) ((b)&DD_CACHEIDXMASK)
#define DD_GETCACHEIDX2(b) (((b)>>DD_CACHEIDXBIT)&DD_CACHEIDXMASK)
#define DD_GETCACHEIDX3(b) (((b)>>(DD_CACHEIDXBIT+DD_CACHEIDXBIT))&DD_CACHEIDXMASK)
#define DD_MAKECACHEIDX(b1,b2,b3) ((b1)|((b2)<<DD_CACHEIDXBIT)|((b3)<<(DD_CACHEIDXBIT+DD_CACHEIDXBIT)))

// maximum texture cache size
#define DD_TEXTUREMAX			128
#if DD_TEXTUREMAX>=DD_CACHEIDXMASK+1
#error
#endif

// automatic texture scaling
#define	DD_TEXTUREDIVIDESTART	(1)
#define	DD_TEXTUREDIVIDEMAX		(8)

#define	DD_GVTXMAX				(4+4+4)
#define	DD_TRIANGLEVERTEXMAX	(128*3)			//nb triangle in draw primitive
#define	DD_LINEVERTEXMAX		(256*2)			//nb line in draw primitive

// default primitive flag (D3DDP_DONOTCLIP|D3DDP_DONOTUPDATEEXTENTS are dummy flags in DX7 see render state (clipping+lighening))
#define	DDEX_PRIMITIVEFLAG (D3DDP_DONOTCLIP|D3DDP_DONOTLIGHT|D3DDP_DONOTUPDATEEXTENTS)

#define DDERROR(E) ddError(E,__LINE__);


#define DD_FONTMAX			8

// vertex format sent to driver

class CddTlVertex
{
public:
	float    m_sx;             /* Screen coordinates */
	float    m_sy;
	float    m_sz;
	float    m_rhw;			/* Reciprocal of homogeneous w */
	union		
	{
		struct 
		{
		D3DCOLOR    m_color;          /* Vertex color */
		} ;
		struct 
		{
			Crgb m_rgbcolor;
		};
	};
	D3DCOLOR    m_specular;       /* Specular component of vertex */
	float    m_tu;             /* Texture coordinates */
	float    m_tv;
	float    m_tu2;             /* Texture coordinates */
	float    m_tv2;
};


class CddTlLineVertex
{
public:
	float    m_sx;             /* Screen coordinates */
	float    m_sy;
	float    m_sz;
	float    m_rhw;			/* Reciprocal of homogeneous w */
	union		
	{
		struct 
		{
		D3DCOLOR    m_color;          /* Vertex color */
		} ;
		struct 
		{
			Crgb m_rgbcolor;
		};
	};
	union		
	{
		struct 
		{
		D3DCOLOR    m_specular;          /* Vertex color */
		} ;
		struct 
		{
			Crgb m_rgbspecular;
		};
	};
};



// mode storing
class CddMode
{
private:
    int m_w;
    int m_h;
    D3DFORMAT	m_format;					// Pixel format in this mode
    DWORD		m_behavior;				// Hardware / Software / Mixed vertex processing
    D3DFORMAT	m_depthstencilformat;		// Which depth/stencil format to use with this mode
public:
	SD getW(void)const {return m_w;};
	SD getH(void)const {return m_h;};
	D3DFORMAT getFormat(void)const {return m_format;};
	D3DFORMAT getDepthStencilFormat(void)const {return m_depthstencilformat;};
	DWORD getBehavior(void) const {return m_behavior;}

	void setW(SD w) {m_w=w;};
	void setH(SD h) {m_h=h;};
	void setFormat(D3DFORMAT format) {m_format=format;};
	void setBehavior(DWORD behavior) {m_behavior=behavior;};
	void setDepthStencilFormat(D3DFORMAT depthstencilformat){m_depthstencilformat=depthstencilformat;};
};

// device storing
class CddDevice
{
public:
    char m_name[40];     
    char m_about[40];    
	char m_desc[40];
    D3DDEVTYPE   m_type;			  // Reference, HAL, etc.
    D3DCAPS8     m_caps;			 // Capabilities of this device
    // Modes for this device
    CddMode		m_mode[DD_MODEMAX];
	int			m_modecount;
	int			m_modecurrent;
	int			m_candowindowed;
	int			m_windowed;
    D3DMULTISAMPLE_TYPE m_multisampletype;
};

// driver storing
class CddDriver 
{
public:
    D3DADAPTER_IDENTIFIER8 m_identifier;		// identifier of driver
    D3DDISPLAYMODE m_desktopdisplaymode;      // Desktop display mode for this adapter
	CddDevice m_device[DD_DEVICEMAX];
    int m_devicecount;
    int m_devicecurrent;
};

// texture cache item
typedef struct 
{
	char filename[_MAX_PATH];
    LPDIRECT3DTEXTURE8	texture;
}
DDTEXTURE;


void getDxVersion(unsigned int *version,unsigned int *platform);


class Cdd
{
public:
	HWND					m_hwnd;			//handle of directdraw window
	LPDIRECT3D8				m_d3d;			// D3D object 
	LPDIRECT3DDEVICE8		m_device;		// D3D device 
	float					m_fwinx1;
	float					m_fwiny1;
	float					m_fwinx2;
	float					m_fwiny2;
	int						m_winx1;
	int						m_winy1;
	int						m_winx2;
	int						m_winy2;


	int						m_fullscreen;
	int						m_w;
	int						m_h;
	int						m_bpp;
	int						m_pitch;

	UD						m_rmask;
	UD						m_gmask;
	UD						m_bmask;
	UD						m_amask;

	UD						m_rshift;
	UD						m_gshift;
	UD						m_bshift;
	UD						m_ashift;

	UD						m_rbitcount;
	UD						m_gbitcount;
	UD						m_bbitcount;
	UD						m_abitcount;

// d3d device info
	CddDriver				m_driver[DD_DRIVERMAX];
	int						m_drivercount;
	int						m_drivercurrent;
	BOOL					m_driverchanged;

//DX8	// texture format info
//DX8		DDTEXTUREFORMAT			m_textureformat[DDBESTTEXTUREFORMAT_MAX];

// zbuffer format info
//DX8		DDZBUFFERFORMAT			m_zbufferformat[DDBESTZBUFFERFORMAT_MAX];
//DX8		int						m_zbufferformatcurrent;

// operations selected for stencil shadow
	int						m_stencilincop;
	int						m_stencildecop;

// texture buffers info
	DDTEXTURE				m_texture[DD_TEXTUREMAX];
	BOOL					m_texture32;

	Cgvtx					*src_v1addr[12];
	Cgvtx					*src_v2addr[12];
	int						src_v1count;
	int						src_v2count;
	Cgvtx					src_vbuffer[12];		// new vertex for 2d clip
	int						src_vbuffercount;

	int						m_renderfilter;
	BOOL					m_renderdither;
	BOOL					m_renderzbuffer;
	BOOL					m_renderantialias;
	BOOL					m_renderfog;
	Crgb					m_renderfogcolor;
	int						m_backbuffercount;
	int						m_trybackbuffercount;

	BNKIDX					m_oldbank;
	SIDEATTR				m_oldattrib;


	int						m_texturedividestart;
	int						m_texturedivide;

	BOOL					m_zbufferenable;		

	BOOL					m_stencilenable;

	float					m_zbuffermax;

	int						m_texturestagemax;
	int						m_usetexturestagemax;

	int						m_rendermipmap;

	// var for triangle tl vertex buffer
	LPDIRECT3DVERTEXBUFFER8	m_trianglevertexbuffer8;
	CddTlVertex				*m_trianglevertex;
	int						m_trianglevertexcount;

	LPDIRECT3DVERTEXBUFFER8	m_linevertexbuffer8;
	CddTlLineVertex			*m_linevertex;
	int						m_linevertexcount;



	// var for stats
	int						m_trianglecount;
	int						m_updaterenderstatecount;
	int						m_setstatecount;
	int						m_settexturestagecount;
	int						m_drawprimitivecount;
	int						m_beginrendercount;

	
	BOOL					m_insetformat;
	RECT					m_hwndrect;
	LONG					m_hwndstyle;
	BOOL					m_novsync;

	float					m_nearplane;	//nearplane clip
	float					m_invnearplane;
	float					m_nearplane_divmaxz;	//nearplane clip
	float					m_invnearplane_divmaxz;


	Cmat					m_matrix;

	Cgfont					m_font[DD_FONTMAX];
	Crgb					m_fontrgb;
//	Crgb					m_fontspecular;

private:
	void ddError(HRESULT err,int line);
	int findDepthStencilFormat(UINT iAdapter,D3DDEVTYPE DeviceType,D3DFORMAT TargetFormat,D3DFORMAT* pDepthStencilFormat);
	int confirmDevice(D3DCAPS8* pCaps,DWORD dwBehavior,D3DFORMAT Format);
	HRESULT buildDeviceList(void);
	int dd_Init(void);
	int dd_BeginRender();
	int dd_EndRender();
	void dd_Polygon(SIDEATTR attrib,Cgvtx **pts,int ptscount,BNKIDX bank);
	void dd_Triangle(SIDEATTR attrib,Cgvtx *a,Cgvtx *b,Cgvtx *c,BNKIDX bank);
	void dd_Line(Cgvtx *a,Cgvtx *b);
	void dd_State(D3DRENDERSTATETYPE state,UD value);
	void dd_SetTextureStageState(DWORD dwstage,D3DTEXTURESTAGESTATETYPE dwstate,DWORD dwvalue);
	void xEdgeCompute(SIDEATTR attrib);
	void yEdgeCompute(SIDEATTR attrib);
	int restoreSurfaces(void);
	void releaseAll(void);
	int countLoadableTexture(void);
	void initRenderState(void);
	void initRenderStage(void);
	void updateRenderBank(SIDEATTR attrib,BNKIDX bank);
	void switchRenderStateReflection(BOOL flag);
	void flushTriangle(void);
	void flushLine(void);
public:
	void flushAll(void);
	void updateRenderState(SIDEATTR attrib);
	BOOL ready(void);
	void clearStats();
	void *lock();
	int unLock(void);
	Cdd();
	~Cdd();
	void setClipRect(int x1,int y1,int x2,int y2);
	void setClipRect(Cv2d const &v1,Cv2d const &v2);

	float getfClipx1(void)const{return m_fwinx1;}
	float getfClipy1(void)const{return m_fwiny1;}
	float getfClipx2(void)const{return m_fwinx2;}
	float getfClipy2(void)const{return m_fwiny2;}

	int getClipx1(void)const{return m_winx1;}
	int getClipy1(void)const{return m_winy1;}
	int getClipx2(void)const{return m_winx2;}
	int getClipy2(void)const{return m_winy2;}

	int getFullScreen(void)const{return(m_fullscreen);}
	//void setfullscreen(int n){m_fullscreen=n;}
	int getW(void)const{return(m_w);}
	int getH(void)const{return(m_h);}

	int getTriangleCount(void)const {return m_trianglecount;}

	int getUpdateRenderStateCount(void)const {return m_updaterenderstatecount;}
	int getDrawPrimitiveCount(void)const {return m_drawprimitivecount;}
	int getSetStateCount(void)const {return m_setstatecount;}
	int getSetTextureStageCount(void)const {return m_settexturestagecount;}

	BOOL setFormat(HWND hwnd,int fullscreen);

	// show and flip
	BOOL flip(void);
	void show(RECT rcsrc,HWND dstwnd,RECT rcdst);

	
	int getBackBufferCount(void)const{return m_backbuffercount;}
	void setBackBufferCount(int count);
	CddDriver *getDriverCurrentAddr(void){return &m_driver[m_drivercurrent];};
	CddDevice *getDeviceCurrentAddr(void){return &(getDriverCurrentAddr()->m_device[getDriverCurrentAddr()->m_devicecurrent]);}
	CddMode *getModeCurrentAddr(void){return &(getDeviceCurrentAddr()->m_mode[getDeviceCurrentAddr()->m_modecurrent]);}

	// video modes
	int getVideoMode(void){return getDeviceCurrentAddr()->m_modecurrent;};
	void setVideoMode(int mode){getDeviceCurrentAddr()->m_modecurrent=mode % getDeviceCurrentAddr()->m_modecount;}
	int searchVideoMode(int w,int h,int bpp);
	int getVideoModeCount(void){return(getDeviceCurrentAddr()->m_modecount);}
	int getVideoModeW(int mode){return(getModeCurrentAddr()->getW());}
	int getVideoModeH(int mode){return(getModeCurrentAddr()->getH());}


	// main screen pixel format
	int getBpp(void)const{return(m_bpp);}
	UD getRmask(void)const{return m_rmask;}
	UD getGmask(void)const{return m_gmask;}
	UD getBmask(void)const{return m_bmask;}
	int getRshift(void)const{return(m_rshift);}
	int getGshift(void)const{return(m_gshift);}
	int getBshift(void)const{return(m_bshift);}
	int getRbitcount(void)const{return(m_rbitcount);}
	int getGbitcount(void)const{return(m_gbitcount);}
	int getBbitcount(void)const{return(m_bbitcount);}
	int getPitch(void)const{return(m_pitch);}

	// clear screen
	void cls(void);
	
	// transform rgb to screen pixel format
	UD rgbTransform(Crgb rgb);

	// fill rectangles
	BOOL fillRectangle(int x1,int y1,int x2,int y2,Crgb color);
	BOOL fillRectangle(int x1,int y1,int x2,int y2,Cuv const &a,Cuv const &b,BNKIDX bank,Crgb const &color,Crgb const &specular);

	// error display
	void error(int dderror,char *msgerror,BOOL stopprogram=TRUE);


	int clear(BOOL clearzbuffer,BOOL cleartarget,BOOL clearstencil,float zbuffer,Crgb color,unsigned int stencil);
	BOOL beginRender(BOOL clearzbuffer,BOOL cleartarget,BOOL clearstencil,float zbuffer,Crgb color,unsigned int stencil);
	void clearRenderState(void);
	void line(Cgvtx a,Cgvtx b);
	void line(float x1,float y1,float x2,float y2);
	void line(Cv2d a,Cv2d b,Crgb color);
	void triangle(SIDEATTR attrib,Cgvtx *a,Cgvtx *b,Cgvtx *c,BNKIDX bank);
	void quad(SIDEATTR attrib,Cgvtx *a,Cgvtx *b,Cgvtx *c,Cgvtx *d,BNKIDX bank);
	void pent(SIDEATTR attrib,Cgvtx *a,Cgvtx *b,Cgvtx *c,Cgvtx *d,Cgvtx *e,BNKIDX bank);
	void endRender(void);
	void stopRender(void);
	BOOL continueRender(void);
	BOOL loadTexture(int n,char *name,BOOL isbump=FALSE);
	void releaseAllTexture(void);
	// main render settings
	int getRenderFilter(void)const{return m_renderfilter;}
	int getRenderMipmap(void)const{return m_rendermipmap;}
	BOOL getRenderDither(void)const{return m_renderdither;}
	BOOL getRenderZbuffer(void)const{return m_renderzbuffer;}
	BOOL getRenderAntialias(void)const{return m_renderantialias;}
	BOOL getRenderFog(void)const{return m_renderfog;}
	Crgb getRenderFogColor(void)const{return m_renderfogcolor;}

	void setRenderFilter(BOOL value,BOOL forcevalue=FALSE);
	void setRenderMipmap(BOOL value,BOOL forcevalue=FALSE);
	void setRenderDither(BOOL value,BOOL forcevalue=FALSE);
	void setRenderZbuffer(BOOL value,BOOL forcevalue=FALSE);
	void setRenderAntialias(BOOL value,BOOL forcevalue=FALSE);
	void setRenderFog(BOOL value,BOOL forcevalue=FALSE);
	void setRenderFogColor(Crgb color,BOOL forcevalue=FALSE);

	// HDC funcs	
	HDC createHdc(void);
	void releaseHdc(HDC hdc);
	BOOL blt(int src_x1,int src_y1,int src_x2,int src_y2,int dst_x1,int dst_y1,int dst_x2,int dst_y2);

	// divide texture setting
	int getCurrentTextureDivide(void)const{return m_texturedivide;}
	void setTextureDividePref(int divide){if(divide<1) divide=1;if(divide>8) divide=8;m_texturedividestart=divide;}
	int getTextureDividePref(void)const{return m_texturedividestart;}
	


	void putPixel(int x,int y,Crgb color,BNKIDX bank);

	//drivers 
	int getDriverCount(void)const {return m_drivercount;} 
	int getDriverCurrent(void)const {return m_drivercurrent;} 
	void setDriverCurrent(int driver);
	char *getDriverName(void){return getDriverCurrentAddr()->m_identifier.Driver;}
	char *getDriverDeviceName(void){return getDeviceCurrentAddr()->m_name;}
	char *getDriverName(int index){return m_driver[index].m_identifier.Driver;}

	char *getDriverDeviceName(int index){return m_driver[index].m_device[m_driver[index].m_devicecurrent].m_name;}
	char *getDriverAbout(void){return m_driver[m_drivercurrent].m_identifier.Description;}
	char *getDriverDeviceAbout(void){return m_driver[m_drivercurrent].m_device[m_driver[m_drivercurrent].m_devicecurrent].m_about;}
	char *getDriverAbout(int index){return m_driver[index].m_identifier.Description;}
	char *getDriverDeviceAbout(int index){return m_driver[index].m_device[m_driver[index].m_devicecurrent].m_about;}

	int	getDriverIsTexturePow2(int index)const
	{
		return 1;
	}
	int	getDriverIsTextureSquareOnly(int index)const
	{
		return 1;
	}
	int	getDriverIsConformant(int index)const
	{
		return 1;
		//return m_driver[index].drivercaps.dwCaps & DDCAPS_3D;
	}
	int	getDriverIs3d(int index)const
	{
		return 1;
		//return m_driver[index].drivercaps.dwCaps & DDCAPS_3D;
	}
	int	getDriverCanRenderWindowed(int index)const
	{
		return 1;
		//return m_driver[index].drivercaps.dwCaps2 & DDCAPS2_CANRENDERWINDOWED;
	}
	int	getDriverIsCertified(int index)const
	{
		return 1;
		//return m_driver[index].drivercaps.dwCaps2 & DDCAPS2_CERTIFIED;
	}
	// stages
	int getTextureStageMax(void)const{return m_texturestagemax;}
	int getUseTextureStageMax(void)const{return m_usetexturestagemax;}
	void setUseTextureStageMax(int val);

	// bank 
	int getBankH(BNKIDX bank)const;
	int getBankW(BNKIDX bank)const;

	// bank possibilities
	int getMinTextureW(void)const;
	int getMaxTextureW(void)const;
	int getMinTextureH(void)const;
	int getMaxTextureH(void)const;

	// memory
	int getFreeTextureMemory(void)const;
	int getTotalTextureMemory(void)const;
	int getLocalFreeTextureMemory(void)const;
	int getLocalTotalTextureMemory(void)const;

	// zbuffer max z
	void setZbufferMax(float maxz)
	{
		m_zbuffermax=maxz;
	}
	float getZbufferMax(void)const{return m_zbuffermax;}


	// to set no vsync in fullscreen mode
	void setNoVsync(BOOL novsync){m_novsync=novsync;}
	BOOL getNoVsync(void)const{return m_novsync;}

	// texture 32 bit
	void setTexture32(BOOL flag){m_texture32=flag;}
	BOOL getTexture32(void){return m_texture32;}

	// stencil shadow funcs
	void drawShadow(void);
	void frontShadow(void);
	void backShadow(void);

	void setNearPlane(float nearplane);
	float getNearPlane(void)const;

	void setCameraMatrix(Cmat const &mat);
	Cmat getCameraMatrix(void);

	Crgb getFontRgb(void)const{return m_fontrgb;}
	void setFontRgb(Crgb const &rgb){m_fontrgb=rgb;}
	int fontPrepare(int id,char *filename,int w,int h,int bank);
	void fontDrawText(int id,SIDEATTR attrib,float x,float y,char *text,float zoom=1.0F);
	void fontDrawTextCenter(int id,SIDEATTR attrib,float x,float y,char *text,float zoom=1.0F);
	float fontGetTextW(int id,char *text,float zoom=1.0F);
	float fontGetTextH(int id,float zoom=1.0F);


	void fontDrawTextFormated(int id,SIDEATTR attrib,float x,float y,char *text,float zoom,float maxw,int align=-1);
	float fontGetTextFormatedH(int id,char *text,float zoom,float maxw);
	float fontGetTextFormatedW(int id,char *text,float zoom,float maxw);
};

#endif
