//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------



#ifndef _G3DEF_H_
#define _G3DEF_H_


#include <math.h>

#define CASSERT_MSG(exp, msg)                                       \
{                                                                   \
    struct cassert_t                                                \
    {                                                               \
        template<bool cond>                                         \
        struct checkCondition {};                                   \
                                                                    \
        template<>                                                  \
        struct checkCondition<true> { enum { CASSERT_##msg }; };    \
    };                                                              \
    enum { CASSERT_##msg };                                         \
    cassert_t::checkCondition<exp>::CASSERT_##msg;                  \
}

#pragma warning( disable : 4711 )
#pragma warning( disable : 4100 )
#pragma warning( disable : 4056 )
#pragma warning( disable : 4201 )


// attrib bits for sides

#define ATTR_TEXTURED			0x00000001		//textured
#define ATTR_GOURAUD			0x00000002		//gouraud
#define ATTR_TRANSPARENT		0x00000004		//transparent
#define ATTR_TRANSLUCIDE		0x00000008		//translucide 
#define ATTR_TRANSLUCIDEMOD		0x00000010		//translucide modulated
#define ATTR_REFLECTION			0x00000020		//reflexion
#define ATTR_BUMP				0x00000040		//bump
#define ATTR_WIRE				0x00000080		//wire


#define ATTR_SELECTED			0x00010000
#define ATTR_LIGHTENED			0x00020000
#define ATTR_NOCOLLID			0x00040000
#define ATTR_ANIMATE1			0x00080000
#define ATTR_ANIMATE2			0x00100000
#define ATTR_ANIMATE3			0x00200000
#define ATTR_ANIMATE4			0x00400000

// concatened attrib

#define ATTR_NEEDSTATE			(ATTR_TEXTURED|ATTR_GOURAUD|ATTR_TRANSPARENT|ATTR_TRANSLUCIDE|ATTR_REFLECTION|ATTR_BUMP|ATTR_WIRE)
#define ATTR_NEEDNORMAL			(ATTR_REFLECTION|ATTR_LIGHTENED)

#define ATTR_ALLTRANSLUCIDE		(ATTR_TRANSLUCIDE|ATTR_TRANSLUCIDEMOD)



#define GETBNKIDX1(b) ((b)&0x7FF)
#define GETBNKIDX2(b) (((b)>>11) & 0x7FF)
#define GETBNKIDX3(b) (((b)>>22) & 0x3FF)


typedef signed short int SW;
typedef unsigned short int UW;
typedef signed int SD;
typedef unsigned int UD;
typedef signed char SB;
typedef unsigned char UB;
typedef unsigned char PIXEL;
typedef unsigned short DPIXEL;
typedef void (*PTRFUNCTION)(void *);
typedef float *PTRFLOAT;

typedef UD				SIDEATTR;
typedef UD				VTXIDX;

typedef unsigned int	BNKIDX;


#define UD_MAX	0xFFFFFFFF
#define UD_MIN  0x00000000
#define SD_MAX	0x7FFFFFFF
#define SD_MIN  0x80000000
#define SB_MAX	0x7F
#define SB_MIN  0x80
#define UB_MAX	0xFF
#define UB_MIN  0x00
#define UW_MAX	0x7FFF
#define UW_MIN  0x8000


#define G3_USEASM					1

#define G3_NBANGLE	4096
#define G3_GRAVITY	981.0F

#ifndef PI
#define PI		3.14159265F
#endif

#define	G3_SHIFT	8

#ifndef TRUE
	#define TRUE	(0==0)
#endif

#ifndef FALSE
	#define FALSE	(0!=0)
#endif


#define MIN(A,B) (A<B ? A : B)
#define MAX(A,B) (A>B ? A : B)

#define MIN2(A,B) (A<B ? A : B)
#define MAX2(A,B) (A>B ? A : B)

#define MAX3(A,B,C) (MAX2(A,MAX2(B,C)))
#define MIN3(A,B,C) (MIN2(A,MIN2(B,C)))

#define MAX4(A,B,C,D) (MAX2(A,MAX2(B,MAX2(C,D))) )
#define MIN4(A,B,C,D) (MIN2(A,MIN2(B,MIN2(C,D))) )


#define BITFIELDSIZE(S) ((S>>5)+1)
#define BITFIELDGET(A,B) (A[B>>5]&(1<<(B & 31)))
#define BITFIELDSET(A,B) {A[B>>5]|=(1<<(B & 31))}
#define BITFIELDCLEAR(A,B) {A[B>>5]&=0xFFFFFFFF-(1<<(B & 31));}


static volatile int randseed=0x89696798;


inline void setintrandseed(int seed)
{
	randseed=seed;
}


inline int intrand(void)
{
  int tmp;
  int rs=randseed;
  tmp=(rs>>16)*16807;
  rs=(rs & 0x0000FFFF)*16807+((tmp & 0x00007FFF)<<16)+(tmp>>15);
  randseed=rs;
  return(rs&0xFFFF);
}




inline int intrand(int minvalue,int maxvalue)
{
	return ((intrand()*(maxvalue-minvalue))>>16)+minvalue;
}

inline float floatrand(float maxvalue)
{
	return (float)(intrand())*(1.0F/65536.0F)*maxvalue;
}

inline float floatclamp(float arg,float min,float max)
{
	if (arg < min)
		return min;
	if (arg > max)
		return max;
	return arg;
}

inline float floatrand(float minvalue,float maxvalue)
{
	return (float)(intrand())*(1.0F/65536.0F)*(maxvalue-minvalue)+minvalue;
}

inline void floatfast(void)
{
	unsigned short fpcontrolword; 
__asm
	{
		fstcw	[fpcontrolword]
		mov		ax,[fpcontrolword]
		and		eax,0x0000fcff				// single precision.
		or		eax,0xc
		mov		[fpcontrolword],ax
		fldcw	[fpcontrolword]
	}
}

inline int float2int(float _f)
{
	int res;
	_asm fld	[_f]
	_asm fistp	[res]
	return res;
}

inline int floattoint(float _f)
{
	int res;
	_asm fld	[_f]
	_asm fistp	[res]
	return res;
}


inline float floatabs(float v)
{
	*((long*)&v) &= ~(1<<31);
	return v;
}

inline float floatmod(float _f,float _div)
{
#if 1
	float v;
	__asm
	{
		fld	[_f]
		fld	[_div]
		fprem
		fstp	[v]
		fstp	st(0)		// magouille pour depiler un truc.
	}
	return v;
#else
	int d=float2int(_f/_div);
	return (_f-(d*_div));
#endif
}

inline float floatmod2pi(float _f)
{
	return floatmod(_f,2.0F*PI);
}

inline float floatamortize(float _f,float _value,float _ratio)
{
	float dec=(_f-_value)*_ratio;
	float ret=_f;
	ret-=dec;
	if(_f>_value)
	{
		if(ret<_value)
			ret=_value;
	}
	else
	{
		if(ret>_value)
			ret=_value;
	}
	return(ret);
}

inline float floatsetinrange(float f,float min,float max)
{
	if(f<min)
		f=min;
	else
	if(f>max)
		f=max;
	return(f);
}

inline float floatsetinabsrange(float f,float v)
{
	if(f<-v)
		f=-v;
	else
		if(f>v)
			f=v;
	return f;
}

inline float floatadjustto(float value,float wantedvalue,float step,float time)
{
	if(value<wantedvalue)
	{
		value+=step*time;
		if(value>wantedvalue)
		{
			value=wantedvalue;
		}
	}
	else
	{
		if(value>wantedvalue)
		{
			value-=step*time;
			if(value<wantedvalue)
			{
				value=wantedvalue;
			}
		}
	}
	return value;
}

#endif