//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#ifndef _G3MAT_H_
#define _G3MAT_H_

#include <math.h>

#include <g3def.h>
#include <g3utils.h>
#include <g3rot.h>
#include <g3splinef.h>
#include <g3simd.h>


class Cv3d;
class Cv4d;
class Cmat;
class Cquat;


inline void KATMAI_prefetcht0(void *addr)
{
	__asm mov eax,[addr]
	PREFETCHT0_EAX
}

inline void KATMAI_prefetcht1(void *addr)
{
	__asm mov eax,[addr]
	PREFETCHT0_EAX
}

inline void KATMAI_prefetcht2(void *addr)
{
	__asm mov eax,[addr]
	PREFETCHT0_EAX
}

inline void KATMAI_prefetchnta(void *addr)
{
	__asm mov eax,[addr]
	PREFETCHNTA_EAX
}

inline void KATMAI_loadmat(Cmat const &mat)
{
	__asm mov eax,[mat]
	MOVUPS_XMM3_EAX
	__asm add eax,12
	MOVUPS_XMM4_EAX
	__asm add eax,12
	MOVUPS_XMM5_EAX
	__asm add eax,12
	MOVUPS_XMM6_EAX
}

inline void KATMAI_load0(Cv3d const &v)
{
	__asm mov eax,[v]
	MOVUPS_XMM0_EAX
}

inline void KATMAI_load1(Cv3d const &v)
{
	__asm mov eax,[v]
	MOVUPS_XMM1_EAX
}

inline void KATMAI_shuffle0(void)
{
	MOVAPS_XMM1_XMM0
	MOVAPS_XMM2_XMM0
	SHUFPS_XMM0_XMM0(0x00)
	SHUFPS_XMM1_XMM1(0x55)
	SHUFPS_XMM2_XMM2(0xAA)
}

inline void KATMAI_shuffle1(void)
{
	MOVAPS_XMM0_XMM1
	MOVAPS_XMM2_XMM1
	SHUFPS_XMM0_XMM0(0x00)
	SHUFPS_XMM1_XMM1(0x55)
	SHUFPS_XMM2_XMM2(0xAA)
}

inline void KATMAI_shufflemulmat(void)
{
	MULPS_XMM0_XMM3
	ADDPS_XMM0_XMM6
	MULPS_XMM1_XMM4
	ADDPS_XMM0_XMM1
	MULPS_XMM2_XMM5
	ADDPS_XMM0_XMM2
}

inline void KATMAI_store(Cv3d *ptrv3d)
{
	__asm mov eax,[ptrv3d]
	MOVUPS_EAX_XMM0
}

inline void KATMAI_add(void)
{
	ADDPS_XMM0_XMM1
}

inline void KATMAI_sub(void)
{
	SUBPS_XMM0_XMM1
}

inline void KATMAI_mul(void)
{
	MULPS_XMM0_XMM1
}




//--------------------------------------------------------------------------------
// CV3D  
//--------------------------------------------------------------------------------
class Cv3d
{
public :
	union
	{
		struct
		{
			float    x;
			float    y;
			float    z;
		};
		float tab[3];
	};


	Cv3d() {}	
	Cv3d(float _xyz) {x=_xyz;y=_xyz;z=_xyz;}	
	void set(float _xyz) {x=_xyz;y=_xyz;z=_xyz;}	
	void set(float _x, float _y, float _z)
	{
		x=_x;
		y=_y;
		z=_z;
	}	
	Cv3d(float _x, float _y, float _z)
	{
		set(_x,_y,_z);
	}	
	Cv3d Cv3d::operator + () const {return (*this);}
	Cv3d Cv3d::operator - () const	{return(Cv3d(-x, -y, -z));}
	Cv3d &operator += (Cv3d const &_v) {x+=_v.x;y+=_v.y;z+=_v.z;return (*this);}
	Cv3d &operator -= (Cv3d const &_v) {x-=_v.x;y-=_v.y;z-=_v.z;return (*this);}
	Cv3d &operator *= (float _k) {x*=_k;y*=_k;z*=_k;return(*this);}
	Cv3d &operator /= (float _k) {float ik=1.0F/_k;x*=ik;y*=ik;z*=ik;return(*this);}
	float len2(void)const {return (float)(x*x+y*y+z*z);}
	float len(void)const {return (float)(sqrt(len2()));}
	int nearly(Cv3d const &_v,float _tolerance);
	void normalize(void);
	void amortize(Cv3d const &value,float ratio,float time);
	void adjustto(Cv3d const &wantedvalue,Cv3d const &step,float time);
	void setinabsrange(Cv3d const &max);
	void setinrange(Cv3d const &min,Cv3d const &max);
	//Cv3d xyzoom(float xzoom,float yzoom) {return Cv3d(x*xzoom,y*yzoom,z);}
	void xyzoom(float xzoom,float yzoom) {x*=xzoom;y*=yzoom;}
	//Cv3d xyzzoom(float xzoom,float yzoom,float zzoom) {return Cv3d(x*xzoom,y*yzoom,z*zzoom);}
	void xyzzoom(float xzoom,float yzoom,float zzoom) {x*=xzoom;y*=yzoom;z*=zzoom;}
	Cv3d &operator ^= (Cv3d const &__v)
	{
		float tempx,tempy,tempz;
		tempx=x;
		tempy=y;
		tempz=z;
		x = tempy*__v.z - tempz*__v.y;
		y = tempz*__v.x - tempx*__v.z;
		z = tempx*__v.y - tempy*__v.x;
		return (*this);
	}
	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);

	float maxxyz(void)const{return MAX(x,MAX(y,z));}
	float minxyz(void)const{return MIN(x,MIN(y,z));}
	

	void rand(Cv3d const &_v);


	void rotatex(float angle);
	void rotatey(float angle);
	void rotatez(float angle);
};

Cv3d spline(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3,Cv3d const &v4,float t);
Cv3d spline(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3,Cv3d const &v4,CsplineFactor const &sf);
Cv3d max3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3);
Cv3d min3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3);
Cv3d max2(Cv3d const &v1,Cv3d const &v2);
Cv3d min2(Cv3d const &v1,Cv3d const &v2);


inline int operator ==(Cv3d const &v3d1, Cv3d const &v3d2) 
{
	return 
		v3d1.x==v3d2.x &&
		v3d1.y==v3d2.y &&
		v3d1.z==v3d2.z;
}

inline void Cv3d::rand(Cv3d const &_v)
{
	x=floatrand(_v.x);
	y=floatrand(_v.y);
	z=floatrand(_v.z);
}

inline int Cv3d::nearly(Cv3d const &_v,float _tolerance)
{
	if(
		floatabs(_v.x-x)<=_tolerance &&
		floatabs(_v.y-y)<=_tolerance &&
		floatabs(_v.z-z)<=_tolerance	
		)
		return(TRUE);
	return(FALSE);
}

inline Cv3d operator + (Cv3d const &_v1, Cv3d const &_v2)
{
	Cv3d res;
	res.x = _v1.x + _v2.x;
	res.y = _v1.y + _v2.y;
	res.z = _v1.z + _v2.z;
	return (res);
}

inline Cv3d operator - (Cv3d const &_v1, Cv3d const &_v2)
{
	Cv3d   res;
	res.x = _v1.x - _v2.x;
	res.y = _v1.y - _v2.y;
	res.z = _v1.z - _v2.z;
	return (res);
}

inline Cv3d operator * (float _k, Cv3d const &_v)
{
	Cv3d   res;
	res.x = _k * _v.x;
	res.y = _k * _v.y;
	res.z = _k * _v.z;
	return (res);
}

inline Cv3d operator * (Cv3d const &_v, float _k)
{
	return (_k*_v);
}

inline Cv3d operator / (Cv3d const &_v, float _k)
{
	_k = 1.0F/_k;
	return _v*_k;
}

inline float operator * (Cv3d const &_v1, Cv3d const &_v2)
{
	return (_v1.x*_v2.x + _v1.y*_v2.y + _v1.z*_v2.z);
}

inline Cv3d operator ^ (Cv3d const &__v1, Cv3d const &__v2)
{
	Cv3d   res;
	res.x = __v1.y*__v2.z - __v1.z*__v2.y;
	res.y = __v1.z*__v2.x - __v1.x*__v2.z;
	res.z = __v1.x*__v2.y - __v1.y*__v2.x;
	return (res);
}


//--------------------------------------------------------------------------------
// CMAT33
//--------------------------------------------------------------------------------

class Cmat33
{
public:
	union
	{
		struct
		{
			float    m11;
			float    m12;
			float    m13;
			float    m21;
			float    m22;
			float    m23;
			float    m31;
			float    m32;
			float    m33;
		};
		struct
		{
			Cv3d	v[3];
		};
	};

	Cmat33() {};
	Cmat33(float _value);
	void set(float _value);
	Cmat33(Crot const &_rot);      // Z X Y
	void set(Crot const &_rot);      // Z X Y
	Cmat33(Cquat const &quater);
	void set(Cquat const &quater);
	Cmat33(Cv3d const &a,Cv3d const &b,Cv3d const &c);
	void set(Cv3d const &a,Cv3d const &b,Cv3d const &c);
	void setrx(float angle);
	void setry(float angle);
	void setrz(float angle);

	Cmat33 &operator += (Cmat33 const &_mat);
	Cmat33 &operator -= (Cmat33 const &_mat);
	Cmat33 &operator *= (float _k);

	Cmat33 inverse(void)const;
	void xyzoom(float _xzoom,float _yzoom);
	void xyzzoom(float _xzoom,float _yzoom,float _zzoom);


	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);
};

Cmat33 operator - (Cmat33 const &_mat1, Cmat33 const &_mat2);
Cmat33 operator * (Cmat33 const &mat1, Cmat33 const &mat2);
Cmat33 operator * (Cmat33 const &mat1, float k);
Cv3d operator * (Cv3d const &v,Cmat33 const &mat);
Cmat33 operator + (Cmat33 const &_mat1, Cmat33 const &_mat2);



//--------------------------------------------------------------------------------
// CMAT
//--------------------------------------------------------------------------------
class Cmat
{
public :
	union
	{
		struct
		{
			float    m11;
			float    m12;
			float    m13;
			float    m21;
			float    m22;
			float    m23;
			float    m31;
			float    m32;
			float    m33;
		};
		struct
		{
			Cmat33   mat33;
		};
	};
	Cv3d	 trs;

	Cmat() {};
	void set();
	Cmat(Cv3d const &_trs);
	void set(Cv3d const &_trs);
	Cmat(float _value);
	void set(float _value);
	Cmat(Crot const &_rot);      // Z X Y
	void set(Crot const &_rot);      // Z X Y
    Cmat(Crot const &_rot,Cv3d const &_trs);
    void set(Crot const &_rot,Cv3d const &_trs);
    Cmat(Cv3d const &_trs1,Crot const &_rot,Cv3d const &_trs);
    void set(Cv3d const &_trs1,Crot const &_rot,Cv3d const &_trs);
	Cmat(Cv3d const &a,Cv3d const &b,Cv3d const &c);
	void set(Cv3d const &a,Cv3d const &b,Cv3d const &c);
	Cmat(Cquat const &quater);
	void set(Cquat const &quater);
	Cmat(Cquat const &_quater,Cv3d const &_trs);
	void set(Cquat const &_quater,Cv3d const &_trs);
	Cmat(Cv3d const &_trs1,Cquat const &_quater,Cv3d const &_trs);
	void set(Cv3d const &_trs1,Cquat const &_quater,Cv3d const &_trs);
	Cmat(Cv3d const &_trs1,Cmat33 const &_mat33,Cv3d const &_trs);
	void set(Cv3d const &_trs1,Cmat33 const &_mat33,Cv3d const &_trs);

	void setrx(float angle);
	void setry(float angle);
	void setrz(float angle);

	Cmat &operator += (Cmat const &_mat);
	Cmat &operator -= (Cmat const &_mat);
	Cmat &operator *= (float _k);

	Cmat inverse(void)const;
	void xyzoom(float _xzoom,float _yzoom);
	void xyzzoom(float _xzoom,float _yzoom,float _zzoom);


	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);
};

Cmat operator - (Cmat const &_mat1, Cmat const &_mat2);
Cmat operator * (Cmat const &mat1, Cmat const &mat2);
Cmat operator * (Cmat const &mat1, Cmat33 const &mat2);
Cmat operator * (Cmat const &mat1, float k);
Cmat operator + (Cmat const &_mat1, Cmat const &_mat2);


//--------------------------------------------------------------------------------
// QUATERNION  
//--------------------------------------------------------------------------------


class Cquat
{
public :
	union
	{
		struct
		{
			float   x;    
			float   y;    
			float   z;    
		};
		struct
		{
			Cv3d	v;
		};
	};
	float   w;    

	void set(){x=y=z=0.0F;w=0.5F;}
	void set(float const _x, float const _y, float const _z, float const _w){x=_x;y=_y;z=_z;w=_w;}
	void set(Cv3d const &_v, float const _w){v=_v;w=_w;}

	Cquat(){}
	Cquat(float const _x, float const _y, float const _z, float const _w) {set(_x,_y,_z,_w);}
	Cquat(Cv3d const &_v, float const _w) {set(_v,_w);}
    void set(Cmat33 const &_mat);
    void set(Cmat const &_mat);
    void set(Crot const &_rot){set(Cmat33(_rot));}

    Cquat(Cmat33 const &_mat){set(_mat);}
    Cquat(Cmat const &_mat) {set(_mat);}
	Cquat(Crot const &_rot){set(_rot);}

    Cquat &operator +=(Cquat const &);
    Cquat &operator -=(Cquat const &);
    Cquat &operator *=(float const);
	Cquat &operator /=(float const);
    Cquat &operator *=(Cquat const &);

    Cquat operator +() const {return Cquat(-x,-y,-z,-w);}
    Cquat operator -() const {return *this;}   

	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);

	void normalize(void);

};

Cquat interpolate(Cquat const &quater1,Cquat const &quater2,float ratio);

//--------------------------------------------------------------------------------
// MIX  
//--------------------------------------------------------------------------------

inline Cv3d operator * (Cv3d const &_v, Cmat const &_mat)
{
	Cv3d   res;
	#if 1//G3_USEASM YA BUG
		_asm
		{
			mov		esi,[_v];
			mov		edi,[_mat];
			fld		[esi]Cv3d.x
			fmul	[edi]Cmat.m11		
			fld		[esi]Cv3d.x
			fmul	[edi]Cmat.m12		// dx
			fld		[esi]Cv3d.x
			fmul	[edi]Cmat.m13		// gx

			fld		[esi]Cv3d.y
			fmul	[edi]Cmat.m21		// by
			fld		[esi]Cv3d.y
			fmul	[edi]Cmat.m22		// ey
			fld		[esi]Cv3d.y
			fmul	[edi]Cmat.m23		// hy
			fxch	st(2)

			faddp	st(5),st(0)
			faddp	st(3),st(0)
			faddp	st(1),st(0)

			fld		[esi]Cv3d.z
			fmul	[edi]Cmat.m31		// cz
			fld		[esi]Cv3d.z
			fmul	[edi]Cmat.m32		// fz
			fld		[esi]Cv3d.z
			fmul	[edi]Cmat.m33		// iz
			fxch	st(2)

			faddp	st(5),st(0)
			faddp	st(3),st(0)
			faddp	st(1),st(0)
			fxch	st(2)
			fadd	[edi]Cmat.trs.x
			fxch	st(1)
			fadd	[edi]Cmat.trs.y
			fxch	st(2)
			fadd	[edi]Cmat.trs.z
			fxch	st(1)
			fstp	[res]Cv3d.x
			fxch	st(1)
			fstp	[res]Cv3d.y
			fstp	[res]Cv3d.z
		}
	#else
		res.x = _v.x*mat.m11 + _v.y*mat.m21 + _v.z*mat.m31 + mat.trs.x;
		res.y = _v.x*mat.m12 + _v.y*mat.m22 + _v.z*mat.m32 + mat.trs.y;
		res.z = _v.x*mat.m13 + _v.y*mat.m23 + _v.z*mat.m33 + mat.trs.z;
	#endif
	return(res);
}

/*inline Cv3d operator % (Cv3d const &_v, Cmat const &mat)
{
	Cv3d   res;
	res.x = _v.x*mat.m11 + _v.y*mat.m21 + _v.z*mat.m31;
	res.y = _v.x*mat.m12 + _v.y*mat.m22 + _v.z*mat.m32;
	res.z = _v.x*mat.m13 + _v.y*mat.m23 + _v.z*mat.m33;
	return(res);
}
*/

inline Cv3d operator / (Cv3d const &_v, Cmat const &mat)
{
	Cv3d   res;
	res.x = _v.x*mat.m11 + _v.y*mat.m12 + _v.z*mat.m13;
	res.y = _v.x*mat.m21 + _v.y*mat.m22 + _v.z*mat.m23;
	res.z = _v.x*mat.m31 + _v.y*mat.m32 + _v.z*mat.m33;
	return (res);
}

Cv3d max3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3);
Cv3d min3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3);
Cv3d normal3(Cv3d const &va,Cv3d const &vb,Cv3d const &vc);


inline Cquat &Cquat::operator +=(Cquat const &quater)
{
	v+=v;
	w+=quater.w;
	return (*this);
}

inline Cquat &Cquat::operator -=(Cquat const &quater)
{
	v-=v;
	w-=quater.w;
	return (*this);
}

inline Cquat &Cquat::operator *=(float const k)
{
    w*=k;
    x*=k;
    y*=k;
    z*=k;
    return (*this);
}

inline Cquat &Cquat::operator /=(float const k)
{
	float ik=1.0F/k;
    w*=ik;
    x*=ik;
    y*=ik;
    z*=ik;
    return (*this);
}

inline Cquat &Cquat::operator *=(Cquat const &quater)
{
    float   tempw, tempx, tempy, tempz;
    tempw = w;
    tempx = x;
    tempy = y;
    tempz = z;
    w=tempw*quater.w-tempx*quater.x-tempy*quater.y-tempz*quater.z;
    x=tempw*quater.x+tempx*quater.w+tempy*quater.z-tempz*quater.y;
    y=tempw*quater.y+tempy*quater.w-tempx*quater.z+tempz*quater.x;
    z=tempw*quater.z+tempz*quater.w+tempx*quater.y-tempy*quater.x;
    return (*this);
}

inline Cquat operator +(Cquat const &quater1, Cquat const &quater2)
{
    Cquat res;
	res.w=quater1.w+quater2.w;
	res.x=quater1.x+quater2.x;
	res.y=quater1.y+quater2.y;
	res.z=quater1.z+quater2.z;
    return (res);
}


inline Cquat operator -(Cquat const &quater1, Cquat const &quater2)
{
    Cquat res;
	res.w=quater1.w-quater2.w;
	res.x=quater1.x-quater2.x;
	res.y=quater1.y-quater2.y;
	res.z=quater1.z-quater2.z;
    return (res);
}


inline Cquat operator *(float const k, Cquat const &quater)
{
    Cquat res;
    res.w=k*quater.w;
    res.x=k*quater.x;
    res.y=k*quater.y;
    res.z=k*quater.z;
    return (res);
}


inline Cquat operator *(Cquat const &quater, float const k)
{
    Cquat      res;
    res.w=quater.w*k;
    res.x=quater.x*k;
    res.y=quater.y*k;
    res.z=quater.z*k;
    return (res);
}                                                                                     

inline Cquat operator *(Cquat const &quater1, Cquat const &quater2)
{
    Cquat res;
    res.w=quater1.w*quater2.w-quater1.x*quater2.x-quater1.y*quater2.y-quater1.z*quater2.z;
    res.x=quater1.w*quater2.x+quater1.x*quater2.w+quater1.y*quater2.z-quater1.z*quater2.y;
    res.y=quater1.w*quater2.y+quater1.y*quater2.w-quater1.x*quater2.z+quater1.z*quater2.x;
    res.z=quater1.w*quater2.z+quater1.z*quater2.w+quater1.x*quater2.y-quater1.y*quater2.x;
    return (res);
}

#endif