//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#ifndef _G3RGB_H_
#define _G3RGB_H_

#include <stdio.h>

#include <g3def.h>
#include <g3file.h>

//--------------------------------------------------------------------------------
// CRGB
//--------------------------------------------------------------------------------

#define MAKERGBA(R, G, B, A)   ( ((((UD)(A)) << 24) | (((UD)(R)) << 16) | (((UD)(G)) << 8) | ((UD)(B))))
#define MAKERGB(R, G, B)       ( ((((UD)(R)) << 16) | (((UD)(G)) << 8) | ((UD)(B))))

class Crgb
{
public:
	union
	{
		UD rgba;
		struct {UB b,g,r,a;};
	};

	Crgb() {}
	Crgb(UD const &_rgba) {rgba=_rgba;}
	Crgb(int const &_rgba) {rgba=_rgba;}
	void set(UD const &_rgba) {rgba=_rgba;}
	Crgb(UB _r,UB _g,UB _b) {rgba=MAKERGBA(_r,_g,_b,255);}
	void set(UB _r,UB _g,UB _b) {rgba=MAKERGBA(_r,_g,_b,255);}
	Crgb(UB _r,UB _g,UB _b,UB _a) {rgba=MAKERGBA(_r,_g,_b,_a);}
	void set(UB _r,UB _g,UB _b,UB _a) {rgba=MAKERGBA(_r,_g,_b,_a);}

	void set(Crgb const &_a,Crgb const &_b,float ratio)
	{
		r=(UB)((int)_a.r+float2int( ((float)_b.r-(float)_a.r)*ratio));
		g=(UB)((int)_a.g+float2int( ((float)_b.g-(float)_a.g)*ratio));
		b=(UB)((int)_a.b+float2int( ((float)_b.b-(float)_a.b)*ratio));
		a=(UB)((int)_a.a+float2int( ((float)_b.a-(float)_a.a)*ratio));
	}

	void set(Crgb const &_a,Crgb const &_b,float aiz,float biz,float z,float ratio)
	{	
		float aa,bb;
		aa=(float)_a.r*aiz;
		bb=(float)_b.r*biz;
		r=(UB)float2int((aa+(bb-aa)*ratio)*z);
		aa=(float)_a.g*aiz;
		bb=(float)_b.g*biz;
		g=(UB)float2int((aa+(bb-aa)*ratio)*z);
		aa=(float)_a.b*aiz;
		bb=(float)_b.b*biz;
		b=(UB)float2int((aa+(bb-aa)*ratio)*z);
		aa=(float)_a.a*aiz;
		bb=(float)_b.a*biz;
		a=(UB)float2int((aa+(bb-aa)*ratio)*z);
	}

	UB getr(void)const{return r;}
	UB getg(void)const{return g;}
	UB getb(void)const{return b;} 
	UB geta(void)const{return a;} 
	UD getrgba(void)const{return rgba;}

	Crgb getrgblight(int light)const
	{
		Crgb res;
		res.r=(UB)(((UD)r*light)>>8);
		res.g=(UB)(((UD)g*light)>>8);
		res.b=(UB)(((UD)b*light)>>8);
		res.a=a;
		return res;
	}

	UW convert444()const
	{
		UD rr=r>>4;
		UD gg=g>>4;
		UD bb=b>>4;
		return (UW)((rr<<8)|(gg<<4)|bb);
	}


	void setr(UB _r) {r=_r;}
	void setg(UB _g) {g=_g;}
	void setb(UB _b) {b=_b;}
	void seta(UB _a) {a=_a;}

	UD light(void) const
	{
		return ((UD)r*19661+(UD)g*38666+(UD)b*7209);
	}
	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);

	Crgb extractspecular(void)const
	{
		Crgb ret;
		if(r>128)
			ret.r=(r-128)<<1;
		else
			ret.r=0;
		if(g>128)
			ret.g=(g-128)<<1;
		else
			ret.g=0;
		if(b>128)
			ret.b=(b-128)<<1;
		else
			ret.b=0;
		ret.a=0;
		return ret;
	}

	Crgb extractmul2(void)const
	{
		Crgb ret;
		if(r<128)
			ret.r=r<<1;
		else
			ret.r=255;
		if(g<128)
			ret.g=g<<1;
		else
			ret.g=255;
		if(b<128)
			ret.b=b<<1;
		else
			ret.b=255;
		ret.a=a;
		return ret;
	}


};

inline Crgb operator + (Crgb const &_a,Crgb const &_b)
{
	Crgb res;
	res.rgba=((_a.rgba&0xFEFEFEFE)>>1)+((_b.rgba&0xFEFEFEFE)>>1);
	return (res);
}

inline Crgb operator * (Crgb const &_a,Crgb const &_b)
{
	Crgb res;
	res.r=(UB)(((UD)_a.r*(UD)_b.r)>>8);
	res.g=(UB)(((UD)_a.g*(UD)_b.g)>>8);
	res.b=(UB)(((UD)_a.b*(UD)_b.b)>>8);
	res.a=(UB)(((UD)_a.a*(UD)_b.a)>>8);
	return (res);
}


inline Crgb operator * (Crgb const &_a,int _b)
{
	Crgb res;
	res.rgba=
		((((_a.rgba>>8)&0x00FF00FF)*_b)&0xFF00FF00)|
		((((_a.rgba & 0x00FF00FF)*_b)>>8)&0x00FF00FF);
	return (res);
}

inline Crgb operator * (int _b,Crgb const &_a)
{
	return _a*_b;
}



//--------------------------------------------------------------------------------
// CFRGB  
//--------------------------------------------------------------------------------
class Cfrgb
{
public :
	union
	{
		struct
		{
			float    r;
			float    g;
			float    b;
			float    a;
		};
		float tab[4];
	};


	Cfrgb() {}	
	void set(float _rgb) {r=_rgb;g=_rgb;b=_rgb;}	
	Cfrgb(float _rgba) {set(_rgba);}	
	void set(float _r, float _g, float _b,float _a)
	{
		r=_r;
		g=_g;
		b=_b;
		a=_a;
	}	
	Cfrgb(float _r, float _g, float _b,float _a)
	{
		set(_r,_g,_b,a);
	}
	void set(float _r, float _g, float _b)
	{
		r=_r;
		g=_g;
		b=_b;
		a=0.0F;
	}	
	Cfrgb(float _r, float _g, float _b)
	{
		set(_r,_g,_b);
	}
	
	Cfrgb Cfrgb::operator + () const {return (*this);}
	Cfrgb Cfrgb::operator - () const	{return(Cfrgb(-r, -g, -b));}
	Cfrgb &operator += (Cfrgb const &_v) {r+=_v.r;g+=_v.g;b+=_v.b;return (*this);}
	Cfrgb &operator -= (Cfrgb const &_v) {r-=_v.r;g-=_v.g;b-=_v.b;return (*this);}
	Cfrgb &operator *= (float _k) {r*=_k;g*=_k;b*=_k;return(*this);}
	Cfrgb &operator /= (float _k) {float ik=1.0F/_k;r*=ik;g*=ik;b*=ik;return(*this);}
	float len2(void)const {return (float)(r*r+g*g+b*b+a*a);}
	float len(void)const {return (float)(sqrt(len2()));}
	int nearly(Cfrgb const &_v,float _tolerance);
	Cfrgb max3(Cfrgb const &v1,Cfrgb const &v2,Cfrgb const &v3);
	void rand(Cfrgb const &_v);
	float light(void) const;
	void lightscale(void)
	{
		r*=0.300F;
		g*=0.600F;
		b*=0.100F;
		r/=172.95F;
		g/=172.95F;
		b/=172.95F;
	}
};

inline int operator ==(Cfrgb const &v3d1, Cfrgb const &v3d2) 
{
	return 
		v3d1.r==v3d2.r &&
		v3d1.g==v3d2.g &&
		v3d1.b==v3d2.b &&
		v3d1.a==v3d2.a;
}

inline void Cfrgb::rand(Cfrgb const &_v)
{
	r=floatrand(_v.r);
	g=floatrand(_v.g);
	b=floatrand(_v.b);
	a=0.0F;
}

inline int Cfrgb::nearly(Cfrgb const &_v,float _tolerance)
{
	if(
		floatabs(_v.r-r)<=_tolerance &&
		floatabs(_v.g-g)<=_tolerance &&
		floatabs(_v.b-b)<=_tolerance &&	
		floatabs(_v.a-a)<=_tolerance	
		)
		return(TRUE);
	return(FALSE);
}

inline Cfrgb operator + (Cfrgb const &_v1, Cfrgb const &_v2)
{
	Cfrgb res;
	res.r = _v1.r + _v2.r;
	res.g = _v1.g + _v2.g;
	res.b = _v1.b + _v2.b;
	res.a = _v1.a + _v2.a;
	return (res);
}

inline Cfrgb operator - (Cfrgb const &_v1, Cfrgb const &_v2)
{
	Cfrgb   res;
	res.r = _v1.r - _v2.r;
	res.g = _v1.g - _v2.g;
	res.b = _v1.b - _v2.b;
	res.a = _v1.a - _v2.a;
	return (res);
}

inline Cfrgb operator * (float _k, Cfrgb const &_v)
{
	Cfrgb   res;
	res.r = _k * _v.r;
	res.g = _k * _v.g;
	res.b = _k * _v.b;
	res.b = _k * _v.a;
	return (res);
}

inline Cfrgb operator * (Cfrgb const &_v, float _k)
{
	return (_k*_v);
}

inline Cfrgb operator / (Cfrgb const &_v, float _k)
{
	_k = 1.0F/_k;
	return _v*_k;
}


inline float Cfrgb::light(void) const
{
	return r*0.300F+g*0.600F+b*0.100F;
}


#endif
