//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#ifndef _G3ROT_H_
#define _G3ROT_H_

#include <g3utils.h>

class Cv3d;
class Cmat;

//--------------------------------------------------------------------------------
// CROT
//--------------------------------------------------------------------------------
class Crot
{
public:
	union
	{
		struct
		{
			float x;
			float y;
			float z;
		};
		float tab[3];
	};

	Crot(){};
	Crot(float _xyz) {x=_xyz;y=_xyz;z=_xyz;}
	void set(float _xyz) {x=_xyz;y=_xyz;z=_xyz;}
	Crot(float _x,float _y,float _z) {x=_x;y=_y;z=_z;}
	void set(float _x,float _y,float _z) {x=_x;y=_y;z=_z;}
	Crot(Cv3d const &a,Cv3d const &b);
	Crot(Cmat const &mat);
	Crot operator - () const {return(Crot(-x, -y, -z));}
	Crot &operator += (Crot const &_r) {x+=_r.x;y+=_r.y;z+=_r.z;return (*this);}
	Crot &operator -= (Crot const &_r) {x-=_r.x;y-=_r.y;z-=_r.z;return (*this);}
	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);
	void rand(Crot const &_rot);
	void centerto0(void);
	void amortize(Crot const &value,float ratio,float time);
	void adjustto(Crot const &wantedvalue,Crot const &step,float time);
	void setinabsrange(Crot const &max);
	void setinrange(Crot const &min,Crot const &max);
};



inline Crot operator + (Crot const &_v1, Crot const &_v2)
{
	Crot rot;
	rot.x = _v1.x + _v2.x;
	rot.y = _v1.y + _v2.y;
	rot.z = _v1.z + _v2.z;
	return (rot);
}

inline Crot operator - (Crot const &_v1, Crot const &_v2)
{
	Crot   rot;
	rot.x = _v1.x - _v2.x;
	rot.y = _v1.y - _v2.y;
	rot.z = _v1.z - _v2.z;
	return (rot);
}

inline Crot operator * (float _k, Crot const &_v)
{
	Crot   rot;
	rot.x = _k * _v.x;
	rot.y = _k * _v.y;
	rot.z = _k * _v.z;
	return (rot);
}

inline Crot operator * (Crot const &_v, float _k)
{
	Crot   rot;
	rot.x = _k * _v.x;
	rot.y = _k * _v.y;
	rot.z = _k * _v.z;
	return (rot);
}

inline Crot operator / (Crot const &_v, float _k)
{
	Crot rot;
	_k = 1.0F/_k;
	rot.x =  _v.x * _k;
	rot.y =  _v.y * _k;
	rot.z =  _v.z * _k;
	return (rot);
}

inline float angle2angle(float a,float b)
{
	while(a>PI) a-=2*PI;
	while(a<-PI) a+=2*PI;
	while(b>PI) b-=2*PI;
	while(b<-PI) b+=2*PI;
	float c=b-a;
	if(c>PI)  c-=2*PI;
	if(c<-PI) c+=2*PI;
	return(c);
}


#endif