//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#ifndef _G3UV_H_
#define _G3UV_H_


#include <g3def.h>
#include <g3splinef.h>
//--------------------------------------------------------------------------------
// CUV
//--------------------------------------------------------------------------------

#define UV_SHIFT	8
#define UV_MAX		16

#include <g3def.h>
#include <g3file.h>

class Cuv
{
public:
	union
	{
		struct
		{
			float u;
			float v;
		};
		struct
		{
			int	  iu;
			int   iv;
		};
		float tab[2];
	};
	
	Cuv() {};

	Cuv(float _value){u=_value;v=_value;}
	void set(float _value){u=_value;v=_value;}
	Cuv(float _u,float _v){u=_u;v=_v;}
	void set(float _u,float _v){u=_u;v=_v;}
//	Cuv(int _iu,int _iv){u=(float)_iu*1.0F/256.0F;v=(float)_iv*1.0F/256.0F;}
//	void set(int _iu,int _iv){u=(float)_iu*1.0F/256.0F;v=(float)_iv*1.0F/256.0F;}

	float getu(void)const {return u;}
	float getv(void)const {return v;}
	void setu(float _u) {u=_u;}
	void setv(float _v) {v=_v;}

	void set(Cuv const &_a,Cuv const &_b,float aiz,float biz,float z,float ratio)
	{	
		float a,b;
		a=_a.u*aiz;
		b=_b.u*biz;
		u=(a+(b-a)*ratio)*z;
		a=_a.v*aiz;
		b=_b.v*biz;
		v=(a+(b-a)*ratio)*z;
	}

	void snap(int w,int h,int snapsize)
	{
		int iu,iv;
		iu=(int)(((u*(float)w)/(float)(snapsize))+0.5F);
		iv=(int)(((v*(float)h)/(float)(snapsize))+0.5F);
		u=(float)((iu*snapsize)/w);
		if(u>=1.0F)
		{
			u=1.0F-((1.0F/w)*0.5F);
		}
		v=(float)((iv*snapsize)/h);
		if(v>=1.0F)
		{
			v=1.0F-((1.0F/h)*0.5F);
		}
	}

	void set(Cuv const &_a,Cuv const &_b,float ratio)
	{	
		u=(_a.u+(_b.u-_a.u)*ratio);
		v=(_a.v+(_b.v-_a.v)*ratio);
	}

	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);

	Cuv operator + () const {return (*this);}
	Cuv operator - () const	{return(Cuv(-u, -v));}
	Cuv &operator += (Cuv const &_v) {u+=_v.u;v+=_v.v;return (*this);}
	Cuv &operator -= (Cuv const &_v) {u-=_v.u;v-=_v.v;return (*this);}
	Cuv &operator *= (float _k) {u*=_k;v*=_k;return(*this);}

/*	Cuv &operator =(Cuv const &s);
	float operator ^ (Cuv const &_v1,Cuv const &_v2);
	Cuv operator + (Cuv const &_v1, Cuv const &_v2);
	Cuv operator - (Cuv const &_v1, Cuv const &_v2);
	Cuv operator * (float _k, Cuv const &_v);
	Cuv operator * (Cuv const &_v, float _k);
	Cuv operator / (Cuv const &_v, float _k);
*/

	BOOL isintriangle(Cuv const &a,Cuv const &b,Cuv const &c)const;

	float len2(void)const{return u*u+v*v;}
	float len(void)const{return (float)sqrt(len2());}

	BOOL iswraporequal(Cuv const &a,float tolerance=1.0F/256.0F)const;
};

/*inline Cuv &Cuv::operator = (Cuv const &s)
{
iu = s.iu;
iv = s.iv;
return (*this);
}*/


inline float operator ^ (Cuv const &_v1,Cuv const &_v2)
 {
	return (
			(_v1.u * _v2.v) - 
			(_v2.u * _v1.v)
			);
 }

inline Cuv operator + (Cuv const &_v1, Cuv const &_v2)
{
	Cuv res;
	res.u = _v1.u + _v2.u;
	res.v = _v1.v + _v2.v;
	return (res);
}

inline Cuv operator - (Cuv const &_v1, Cuv const &_v2)
{
	Cuv   res;
	res.u = _v1.u - _v2.u;
	res.v = _v1.v - _v2.v;
	return (res);
}

inline Cuv operator * (float _k, Cuv const &_v)
{
	Cuv   res;
	res.u = _k * _v.u;
	res.v = _k * _v.v;
	return (res);
}

inline Cuv operator * (Cuv const &_v, float _k)
{
	Cuv   res;
	res.u = _k * _v.u;
	res.v = _k * _v.v;
	return (res);
}

inline Cuv operator / (Cuv const &_v, float _k)
{
	Cuv res;
	float ik=1.0F/_k;
	res.u =  _v.u * ik;
	res.v =  _v.v * ik;
	return (res);
}

inline float operator * (Cuv const &_v1, Cuv const &_v2)
{
	return (_v1.u*_v2.u + _v1.v*_v2.v);
}

inline float len2(Cuv const &_v1, Cuv const &_v2)
{
	return (_v1.u-_v2.u)*(_v1.u-_v2.u)+(_v1.v-_v2.v)*(_v1.v-_v2.v);
}

inline int operator ==(Cuv const &uv1, Cuv const &uv2) 
{
	return(
		uv1.u==uv2.u &&
		uv1.v==uv2.v);
}

Cuv getpos(
		   float x,float y,
		   float x1,float y1,float iz1,Cuv uv1,
		   float x2,float y2,float iz2,Cuv uv2,
		   float x3,float y3,float iz3,Cuv uv3
		   );


inline Cuv spline(Cuv const &v1,Cuv const &v2,Cuv const &v3,Cuv const &v4,CsplineFactor const &sf)
{
	Cuv res;
	res.u=v1.u*sf.getm1()+v2.u*sf.getm2()+v3.u*sf.getm3()+v4.u*sf.getm4();
	res.v=v1.v*sf.getm1()+v2.v*sf.getm2()+v3.v*sf.getm3()+v4.v*sf.getm4();
	return(res);
}

inline Cuv spline(Cuv const &v1,Cuv const &v2,Cuv const &v3,Cuv const &v4,float t)
{
	Cuv res;
	res=spline(v1,v2,v3,v4,CsplineFactor(t));
	return(res);
}


#endif
