//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#ifndef _G3V2D_H_
#define _G3V2D_H_

#include <g3def.h>
#include <g3file.h>

//--------------------------------------------------------------------------------
// Cv2d
//--------------------------------------------------------------------------------
class Cv2d
{
public:
	union
	{
		struct
		{
			float    x;
			float    y;
		};
		float tab[2];
	};
public:
	Cv2d() {}
	Cv2d(float _x, float _y) {x=_x;y=_y;}	
	void set(float _x, float _y) {x=_x;y=_y;}	
	Cv2d(float _xy) {x=_xy;y=_xy;}	
	void set(float _xy) {x=_xy;y=_xy;}	

	Cv2d Cv2d::operator + () const {return (*this);}
	Cv2d Cv2d::operator - () const	{return(Cv2d(-x, -y));}
	Cv2d &operator += (Cv2d const &_v) {x+=_v.x;y+=_v.y;return (*this);}
	Cv2d &operator -= (Cv2d const &_v) {x-=_v.x;y-=_v.y;return (*this);}
	Cv2d &operator *= (float _k) {x*=_k;y*=_k;return(*this);}

	float getx(void)const{return x;}
	float gety(void)const{return y;}
	float len2(void)const{return x*x+y*y;}
	float len(void)const{return (float)sqrt(len2());}

	BOOL isintriangle(Cv2d const &a,Cv2d const &b,Cv2d const &c)const;

	int fread(G3FILE *handle);
	int fwrite(G3FILE *handle);

};

inline float operator ^ (Cv2d const &_v1,Cv2d const &_v2)
 {
	return (
			(_v1.x * _v2.y) - 
			(_v2.x * _v1.y)
			);
 }

inline Cv2d operator + (Cv2d const &_v1, Cv2d const &_v2)
{
	Cv2d res;
	res.x = _v1.x + _v2.x;
	res.y = _v1.y + _v2.y;
	return (res);
}

inline Cv2d operator - (Cv2d const &_v1, Cv2d const &_v2)
{
	Cv2d   res;
	res.x = _v1.x - _v2.x;
	res.y = _v1.y - _v2.y;
	return (res);
}

inline Cv2d operator * (float _k, Cv2d const &_v)
{
	Cv2d   res;
	res.x = _k * _v.x;
	res.y = _k * _v.y;
	return (res);
}

inline Cv2d operator * (Cv2d const &_v, float _k)
{
	Cv2d   res;
	res.x = _k * _v.x;
	res.y = _k * _v.y;
	return (res);
}

inline Cv2d operator / (Cv2d const &_v, float _k)
{
	Cv2d res;
	float ik=1.0F/_k;
	res.x =  _v.x * ik;
	res.y =  _v.y * ik;
	return (res);
}

inline float operator * (Cv2d const &_v1, Cv2d const &_v2)
{
	return (_v1.x*_v2.x + _v1.y*_v2.y);
}

inline float len2(Cv2d const &_v1, Cv2d const &_v2)
{
	return (_v1.x-_v2.x)*(_v1.x-_v2.x)+(_v1.y-_v2.y)*(_v1.y-_v2.y);
}

#endif