//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
//	comment : tga file
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#include <stdafx.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include <g3def.h>
#include <g3tga.h>

#pragma pack( push)
#pragma pack(1)

typedef struct
{
  UB  fabric;
  UB  version;
  UB  compression;
  UB  nbplanes;
  UW  xmini;
  UW  ymini;
  UW  xmaxi;
  UW  ymaxi;
  UB  dummy[128-3*4];
}
pcxheader_t;
#pragma pack(pop)


#pragma pack(push,1)
typedef struct
{
	unsigned char numid;	
	unsigned char maptyp;
	unsigned char imgtyp;	
	short maporig;
	short mapsize;
	unsigned char mapbits;
	short xorig;
	short yorig;
	short xsize;
	short ysize;
	unsigned char pixsize;
	unsigned char imgdes;
} tgaheader_t;
#pragma pack(pop)

Ctga::Ctga()
{
	m_w=0;
	m_h=0;
	m_bitdepth=0;
	m_data=NULL;
	m_palette=NULL;
}

Ctga::~Ctga()
{
	release();
}


void Ctga::release(void)
{
	if(m_data!=NULL)
	{
		delete [] m_data;
		m_data=NULL;
	}
	if(m_palette!=NULL)
	{
		delete [] m_palette;
		m_palette=NULL;
	}
	m_bitdepth=0;
	m_w=0;
	m_h=0;
}

int Ctga::getW(void)const
{
	return m_w;
}

int Ctga::getH(void)const
{
	return m_h;
}

int Ctga::getBitDepth(void)const
{
	return m_bitdepth;
}

Crgb *Ctga::getPalette(void)const
{
	return m_palette;
}

Crgb Ctga::getRgb(int x,int y)const
{
	UB *addr;
	_ASSERT(m_data);
	_ASSERT(m_bitdepth==24 || m_bitdepth==32);
	addr=m_data;
	if(m_bitdepth==24)
	{
		addr+=(y*m_w+x)*3;
	}
	else
	{
		addr+=((y*m_w+x)<<2);
	}
	return(Crgb());
}


UB *Ctga::getData(void)const
{
	return(m_data);
}

BOOL Ctga::save(char *filename)
{
	G3FILE *handle;
	char name[_MAX_PATH];
	_ASSERT(m_bitdepth==24 || m_bitdepth==32);
	fextput(filename,".TGA",name);
	handle=g3fopen(name,"wb");
	if(handle!=NULL)
	{
		tgaheader_t header;
		header.numid=0;	
		header.maptyp=0;
		header.imgtyp=0x2;	
		header.maporig=0;
		header.mapsize=0;
		header.mapbits=0;
		header.xorig=0;
		header.yorig=0;
		header.xsize=(short)m_w;
		header.ysize=(short)m_h;
		header.pixsize=(UB)m_bitdepth;
		header.imgdes=0x8;
		g3fwrite(&header,sizeof(header),1,handle);
		_vreverse();
		g3fwrite(m_data,m_w*m_h*(m_bitdepth>>3),1,handle);
		_vreverse();
		g3fclose(handle);
		return TRUE ;
	}
	return FALSE;
}


void Ctga::putPixel(int x,int y,Crgb color)
{
	_ASSERT(m_bitdepth==24 || m_bitdepth==32);
	_ASSERT(m_data);
	if(
		x>=0 && x<m_w &&
		y>=0 && y<m_h 
		)
	{
		UB *ptr;
		ptr=m_data+(m_bitdepth>>3)*(m_w*y+x);
		ptr[0]=color.r;
		ptr[1]=color.g;
		ptr[2]=color.b;
	}
}

BOOL Ctga::load(char *filename)
{
	G3FILE *handle;
	char name[_MAX_PATH];

	release();

/*	//------------------- read a PCX file ------------------------------
	fextput(filename,".PCX",name);
	handle=g3fopen(name,"rb");
	if(handle!=NULL)
	{
		pcxheader_t pcxheader;
		int size;
		int i;
		UB *cdata,pixel;
		int count;
		m_bitdepth = 8;
		g3fread(&pcxheader,sizeof(pcxheader),1,handle);
		m_w=(SD)((UD)(pcxheader.xmaxi-pcxheader.xmini+1));
		m_h=(SD)((UD)(pcxheader.ymaxi-pcxheader.ymini+1));
		size=m_w*m_h*sizeof(UB);
		cdata=m_data=new UB[size];
		_ASSERT(m_data);
		// read and decompress the picture
		while(size>0)
		{
			g3fread(&pixel,sizeof(pixel),1,handle);
			if(pixel<=0x0C0)
			{
				*cdata=pixel;
				cdata++;
				size--;
			}
			else
			{
				count=(int)pixel-0x00C0;
				g3fread(&pixel,sizeof(pixel),1,handle);
				size-=(SD)((UD)count);
				do
				{
					*cdata=pixel;
					cdata++;
					count--;
				}
				while(count);
			}
		}
		// verify format
		g3fread(&pixel,sizeof(pixel),1,handle);
		// read the palette
		m_palette=new Crgb[256];
		_ASSERT(m_palette);
		for(i=0;i<256;i++)
		{
			UB r,g,b;
			g3fread(&r,sizeof(r),1,handle);
			g3fread(&g,sizeof(g),1,handle);
			g3fread(&b,sizeof(b),1,handle);
			m_palette[i]=Crgb(r,g,b);
		}
		g3fclose(handle);
		return(TRUE);
	}
	*/
	//------------------- read a TGA file ------------------------------
	fextput(filename,".TGA",name);
	handle=g3fopen(name,"rb");
	if(handle!=NULL)
	{
		tgaheader_t header;
		g3fread(&header,sizeof(header),1,handle);
		if ((header.numid==0)	&&
			(header.maptyp==0)	&&	//,"TGA-File must be no-paletted.");
			(header.imgtyp==2)	&&	//,"TGA-File imgtyp not supported.");
			(header.mapbits==0) 	//,"TGA-File must be no-paletted.");
			)
		{	// Load a 24 bits TGA.
			if(header.pixsize==24)
			{
				m_bitdepth = 24;
				m_w=header.xsize;
				m_h=header.ysize;
				//fseek(handle,header.imgdes,SEEK_CUR);	// Seek le descripteur d'image. (commentaire)
				m_data = new UB[m_w*m_h*3];
				_ASSERT(m_data);//,"NEW error.");
				g3fread(m_data,1,m_w*m_h*3,handle);
			}
			// Load a 32 bits TGA
			else if(header.pixsize==32)
			{
				m_bitdepth = 32;
				m_w=header.xsize;
				m_h=header.ysize;
				//fseek(handle,header.imgdes,SEEK_CUR);	// Seek le descripteur d'image. (commentaire)
				m_data = new UB[m_w*m_h*4];
				_ASSERT(m_data);//,"NEW error.");
				g3fread(m_data,1,m_w*m_h*4,handle);
			}
			_vreverse();
			g3fclose(handle);
			return(TRUE);
		}
		g3fclose(handle);
	}
	return(FALSE);
}


BOOL Ctga::getInfoFromFile(char *filename,int *pw,int *ph,int *pbpp)
{
	G3FILE *handle;
	char name[_MAX_PATH];
	//------------------- read a TGA file ------------------------------
	fextput(filename,".TGA",name);
	handle=g3fopen(name,"rb");
	if(handle!=NULL)
	{
		tgaheader_t header;
		g3fread(&header,sizeof(header),1,handle);
		if ((header.numid==0)	&&
			(header.maptyp==0)	&&	//,"TGA-File must be no-paletted.");
			(header.imgtyp==2)	&&	//,"TGA-File imgtyp not supported.");
			(header.mapbits==0) 	//,"TGA-File must be no-paletted.");
			)
		{
			if(pw)
			{
				*pw=(int)header.xsize;
			}
			if(ph)
			{
				*ph=(int)header.ysize;
			}
			if(pbpp)
			{
				*pbpp=(int)header.pixsize;
			}
			g3fclose(handle);
			return(TRUE);
		}
		g3fclose(handle);
	}
	return(FALSE);
}


void Ctga::_vreverse(void)
{
	if(!m_data) return;
	int nbline = (m_h>>1);
	if (nbline>0)
	{
		int	linesize = m_w*(m_bitdepth>>3);
		UB *p1 = m_data;
		UB *p2 = m_data + (m_h-1)*linesize;
		do
		{
			memswap(p1,p2,linesize);
			p1 += linesize;
			p2 -= linesize;
			nbline--;
		}
		while (nbline);
	}
}





void Ctga::convertRgba16(UD rmask,UD gmask,UD bmask,UD amask)
{
	_ASSERT(m_bitdepth==24 || m_bitdepth==32 || m_bitdepth==8);
	UW *newdata=new UW[m_w*m_h];
	_ASSERT(newdata);
	UB *srcptr;
	UW *dstptr;
	srcptr=(UB *)m_data;
	dstptr=newdata;

	m_rmask=rmask;
	m_gmask=gmask;
	m_bmask=bmask;
	m_amask=amask;

	UD rshift=getrbitshift(rmask);
	UD gshift=getrbitshift(gmask);
	UD bshift=getrbitshift(bmask);
	UD ashift=getrbitshift(amask);
	UD rbitcount=getbitcount(rmask);
	UD gbitcount=getbitcount(gmask);
	UD bbitcount=getbitcount(bmask);
	UD abitcount=getbitcount(amask);
	
	switch(m_bitdepth)
	{
	case 8:
		{
			_ASSERT(m_palette);
			for(int i=0;i<m_w*m_h;i++)
			{
				UD r,g,b,a;
				b=m_palette[*srcptr].getb();
				g=m_palette[*srcptr].getg();
				r=m_palette[*srcptr].getr();
				a=0;
				if(*srcptr) a=0xFFFFFFFF;
				r>>=(8-rbitcount);
				g>>=(8-gbitcount);
				b>>=(8-bbitcount);
				a>>=(8-abitcount);
				r<<=rshift;
				g<<=gshift;
				b<<=bshift;
				a<<=ashift;
				r&=rmask;
				g&=gmask;
				b&=bmask;
				a&=amask;
				*dstptr=(UW)(r|g|b|a);
				srcptr++;
				dstptr++;
			}
		}
		break;
	case 24:
		{
			if(amask)
			{
				for(int i=0;i<m_w*m_h;i++)
				{
					UD r,g,b,a;
					b=srcptr[0];
					g=srcptr[1];
					r=srcptr[2];
					if(r|g|b)
					{
						a=255;
					}
					else
					{
						a=0;
					}
					r>>=(8-rbitcount);
					g>>=(8-gbitcount);
					b>>=(8-bbitcount);
					a>>=(8-abitcount);
					r<<=rshift;
					g<<=gshift;
					b<<=bshift;
					a<<=ashift;
					r&=rmask;
					g&=gmask;
					b&=bmask;
					a&=amask;
					*dstptr=(UW)(r|g|b|a);
					srcptr+=3;
					dstptr++;
				}
			}
			else
			{
				for(int i=0;i<m_w*m_h;i++)
				{
					UD r,g,b;
					b=srcptr[0];
					g=srcptr[1];
					r=srcptr[2];
					r>>=(8-rbitcount);
					g>>=(8-gbitcount);
					b>>=(8-bbitcount);
					r<<=rshift;
					g<<=gshift;
					b<<=bshift;
					r&=rmask;
					g&=gmask;
					b&=bmask;
					*dstptr=(UW)(r|g|b);
					srcptr+=3;
					dstptr++;
				}
			}
		}
		break;
	case 32:
		{
			for(int i=0;i<m_w*m_h;i++)
			{
				UD r,g,b,a;
				b=srcptr[0];
				g=srcptr[1];
				r=srcptr[2];
				a=srcptr[3];
				r>>=(8-rbitcount);
				g>>=(8-gbitcount);
				b>>=(8-bbitcount);
				a>>=(8-abitcount);
				r<<=rshift;
				g<<=gshift;
				b<<=bshift;
				a<<=ashift;
				r&=rmask;
				g&=gmask;
				b&=bmask;
				a&=amask;
				*dstptr=(UW)(r|g|b|a);
				srcptr+=4;
				dstptr++;
			}
		}
		break;

	}
	delete [] m_data;
	m_data=(UB *)newdata;
	m_bitdepth=16;
}


void Ctga::convertRgba32(UD rmask,UD gmask,UD bmask,UD amask)
{
	_ASSERT(m_bitdepth==24 || m_bitdepth==32);
	UD *newdata=new  UD[m_w*m_h];
	_ASSERT(newdata);
	UB *srcptr;
	UD *dstptr;
	srcptr=(UB *)m_data;
	dstptr=newdata;
	m_rmask=rmask;
	m_gmask=gmask;
	m_bmask=bmask;
	m_amask=amask;
	UD rshift=getrbitshift(rmask);
	UD gshift=getrbitshift(gmask);
	UD bshift=getrbitshift(bmask);
	UD ashift=getrbitshift(amask);
	UD rbitcount=getbitcount(rmask);
	UD gbitcount=getbitcount(gmask);
	UD bbitcount=getbitcount(bmask);
	UD abitcount=getbitcount(amask);
	switch(m_bitdepth)
	{
	case 24:
		{
			if(amask)
			{
				for(int i=0;i<m_w*m_h;i++)
				{
					UD r,g,b,a;
					b=srcptr[0];
					g=srcptr[1];
					r=srcptr[2];
					if(r|g|b)
					{
						a=255;
					}
					else
					{
						a=0;
					}
					r>>=(8-rbitcount);
					g>>=(8-gbitcount);
					b>>=(8-bbitcount);
					a>>=(8-abitcount);
					r<<=rshift;
					g<<=gshift;
					b<<=bshift;
					a<<=ashift;
					r&=rmask;
					g&=gmask;
					b&=bmask;
					a&=amask;
					*dstptr=(UD)(r|g|b|a);
					srcptr+=3;
					dstptr++;
				}
			}
			else
			{
				for(int i=0;i<m_w*m_h;i++)
				{
					UD r,g,b;
					b=srcptr[0];
					g=srcptr[1];
					r=srcptr[2];
					r>>=(8-rbitcount);
					g>>=(8-gbitcount);
					b>>=(8-bbitcount);
					r<<=rshift;
					g<<=gshift;
					b<<=bshift;
					r&=rmask;
					g&=gmask;
					b&=bmask;
					*dstptr=(UD)(r|g|b);
					srcptr+=3;
					dstptr++;
				}
			}
		}
		break;
	case 32:
		{
			for(int i=0;i<m_w*m_h;i++)
			{
				UD r,g,b,a;
				b=srcptr[0];
				g=srcptr[1];
				r=srcptr[2];
				a=srcptr[3];
				r>>=(8-rbitcount);
				g>>=(8-gbitcount);
				b>>=(8-bbitcount);
				a>>=(8-abitcount);
				r<<=rshift;
				g<<=gshift;
				b<<=bshift;
				a<<=ashift;
				r&=rmask;
				g&=gmask;
				b&=bmask;
				a&=amask;
				*dstptr=(UD)(r|g|b|a);
				srcptr+=4;
				dstptr++;
			}
		}
		break;

	}
	delete [] m_data;
	m_data=(UB *)newdata;
	m_bitdepth=32;
}


void Ctga::convertUvl16(UD umask,UD vmask,UD lmask)
{
	_ASSERT(m_bitdepth==24 || m_bitdepth==32 || m_bitdepth==8);
	UW *newdata=new unsigned short[m_w*m_h*sizeof(UW)];
	_ASSERT(newdata);
	UB *srcptr;
	UW *dstptr;
	srcptr=(UB *)m_data;
	dstptr=newdata;

	m_umask=umask;
	m_vmask=vmask;
	m_lmask=lmask;
	m_amask=0;

	UD ushift=getrbitshift(umask);
	UD vshift=getrbitshift(vmask);
	UD lshift=getrbitshift(lmask);
	UD ubitcount=getbitcount(umask);
	UD vbitcount=getbitcount(vmask);
	UD lbitcount=getbitcount(lmask);


	switch(m_bitdepth)
	{
	case 8:
		{
		}
		break;
	case 24:
		{
			for(int y=0;y<m_h;y++)
			{
				for(int x=0;x<m_h;x++)
				{
					int u,v,l;
					u=srcptr[(y*m_w+x)*3]-srcptr[(y*m_w+((x+1)%m_w))*3];
					v=srcptr[(y*m_w+x)*3]-srcptr[(((y+1)%m_h)*m_w+x)*3];
					l=srcptr[(y*m_w+x)*3+1];
					u>>=(8-ubitcount);
					v>>=(8-vbitcount);
					l>>=(8-lbitcount);
					u<<=ushift;
					v<<=vshift;
					l<<=lshift;
					u&=umask;
					v&=vmask;
					l&=lmask;
					*dstptr=(UW)(u|v|l);
					dstptr++;
				}
			}
		}
		break;
	case 32:
		{
		{
			for(int y=0;y<m_h;y++)
			{
				for(int x=0;x<m_h;x++)
				{
					int u,v,l;
					u=srcptr[(y*m_w+x)*4]-srcptr[(y*m_w+((x+1)%m_w))*4];
					v=srcptr[(y*m_w+x)*4]-srcptr[(((y+1)%m_h)*m_w+x)*4];
					l=srcptr[(y*m_w+x)*4+1];
					u>>=(8-ubitcount);
					v>>=(8-vbitcount);
					l>>=(8-lbitcount);
					u<<=ushift;
					v<<=vshift;
					l<<=lshift;
					u&=umask;
					v&=vmask;
					l&=lmask;
					*dstptr=(UW)(u|v|l);
					dstptr++;
				}
			}
		}
		}
		break;

	}
	delete [] m_data;
	m_data=(UB *)newdata;
}


void Ctga::setSize(int neww,int newh)
{
	_ASSERT(m_bitdepth==16 || m_bitdepth==24 || m_bitdepth==32);
	_ASSERT(m_data);
	switch(m_bitdepth)
	{
	case 24:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[3*neww*newh];
			_ASSERT(newdata);
			int stepy=(m_h<<16)/newh;
			int stepx=(m_w<<16)/neww;
			int gety=0;
			for(int y=0;y<newh;y++)
			{
				int getx=0;
				for(int x=0;x<neww;x++)
				{
					UB *getaddr=m_data+( ((gety>>16)*m_w+(getx>>16)) *3 );
					cnewdata[0]=getaddr[0];
					cnewdata[1]=getaddr[1];
					cnewdata[2]=getaddr[2];
					cnewdata+=3;
					getx+=stepx;
				}
				gety+=stepy;
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	case 32:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[4*neww*newh];
			_ASSERT(newdata);
			int stepy=(m_h<<16)/newh;
			int stepx=(m_w<<16)/neww;
			int gety=0;
			for(int y=0;y<newh;y++)
			{
				int getx=0;
				for(int x=0;x<neww;x++)
				{
					UB *getaddr=m_data+( ((gety>>16)*m_w+(getx>>16)) * 4 );
					cnewdata[0]=getaddr[0];
					cnewdata[1]=getaddr[1];
					cnewdata[2]=getaddr[2];
					cnewdata+=4;
					getx+=stepx;
				}
				gety+=stepy;
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	}
	m_w=neww;
	m_h=newh;
}

void Ctga::divn(int n)
{
	_ASSERT(m_bitdepth==16 || m_bitdepth==24 || m_bitdepth==32);
	_ASSERT(m_data);
	switch(m_bitdepth)
	{
	case 16:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[2*(m_w/n)*(m_h/n)];
			_ASSERT(newdata);
			for(int sy=0;sy<(m_h/n);sy++)
			{
				for(int sx=0;sx<(m_w/n);sx++)
				{
					UD r,g,b,a;
					r=g=b=a=0;
					for(int ssy=sy*n;ssy<(sy+1)*n;ssy++)
					{
						for(int ssx=sx*n;ssx<(sx+1)*n;ssx++)
						{
							UW *ptr=(UW *)m_data+(ssy)*(m_w)+(ssx);
							r+=((UD)(*(ptr))&m_rmask);
							g+=((UD)(*(ptr))&m_gmask);
							b+=((UD)(*(ptr))&m_bmask);
							a+=((UD)(*(ptr))&m_amask);
						}
					}
					r/=(n*n);
					g/=(n*n);
					b/=(n*n);
					a/=(n*n);
					*((UW *)cnewdata)=(UW)((r & m_rmask)|(g & m_gmask)|(b & m_bmask)|(a & m_amask));
					cnewdata+=2;
				}
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	case 24:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[3*(m_w/n)*(m_h/n)];
			_ASSERT(newdata);
			int linesize=(m_w*3);
			for(int y=0;y<(m_h/n);y++)
			{
				for(int x=0;x<(m_w/n);x++)
				{
					UB *ptr;
					UD r,g,b;
					ptr=m_data+(y*n)*linesize+(x*n)*3;
					r=g=b=0;
					for(int dy=0;dy<n;dy++)
					{
						for(int dx=0;dx<n;dx++)
						{
							UB *ptr2=ptr+dy*linesize+dx*3;
							r+=*(ptr2+0);
							g+=*(ptr2+1);
							b+=*(ptr2+2);
						}
					}
					r/=(n*n);
					g/=(n*n);
					b/=(n*n);
					*cnewdata++=(UB)r;
					*cnewdata++=(UB)g;
					*cnewdata++=(UB)b;
				}
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	case 32:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[4*(m_w/n)*(m_h/n)];
			_ASSERT(newdata);
			int linesize=(m_w*4);
			for(int y=0;y<(m_h/n);y++)
			{
				for(int x=0;x<(m_w/n);x++)
				{
					UB *ptr;
					UD r,g,b,a;
					ptr=m_data+(y*n)*linesize+(x*n)*4;
					r=0;
					g=0;
					b=0;
					a=0;
					for(int dy=0;dy<n;dy++)
					for(int dx=0;dx<n;dx++)
					{
						UB *ptr2=ptr+dy*linesize+dx*4;
						r+=*(ptr2+0);
						g+=*(ptr2+1);
						b+=*(ptr2+2);
						a+=*(ptr2+3);
					}
					r/=(n*n);
					g/=(n*n);
					b/=(n*n);
					a/=(n*n);
					*cnewdata++=(UB)r;
					*cnewdata++=(UB)g;
					*cnewdata++=(UB)b;
					*cnewdata++=(UB)a;
				}
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	}
	m_w/=n;
	m_h/=n;
}

void Ctga::divnSigned(int n)
{
	_ASSERT(m_bitdepth==16 || m_bitdepth==24 || m_bitdepth==32);
	_ASSERT(m_data);

	UD l_rshift=getlbitshift(m_rmask);
	UD l_gshift=getlbitshift(m_gmask);
	UD l_bshift=getlbitshift(m_bmask);
	UD l_ashift=getlbitshift(m_amask);

	switch(m_bitdepth)
	{
	case 16:
		{
			UB *cnewdata,*newdata;
			cnewdata=newdata=new UB[2*(m_w/n)*(m_h/n)];
			_ASSERT(newdata);
			for(int sy=0;sy<(m_h/n);sy++)
			{
				for(int sx=0;sx<(m_w/n);sx++)
				{
					int r,g,b,a;
					r=g=b=a=0;
					for(int ssy=sy*n;ssy<(sy+1)*n;ssy++)
					{
						for(int ssx=sx*n;ssx<(sx+1)*n;ssx++)
						{
							UW *ptr=(UW *)m_data+(ssy)*(m_w)+(ssx);

							SW rr=((UW)((*(ptr)&m_rmask)<<l_rshift))>>(l_rshift);
							SW gg=((UW)((*(ptr)&m_gmask)<<l_gshift))>>(l_gshift);
							SW bb=((UW)((*(ptr)&m_bmask)<<l_bshift))>>(l_bshift);
							SW aa=((UW)((*(ptr)&m_amask)<<l_ashift))>>(l_ashift);

							r+=((*(ptr))&m_rmask);
							g+=((*(ptr))&m_gmask);
							b+=((*(ptr))&m_bmask);
							a+=((*(ptr))&m_amask);
						}
					}
					r/=(n*n);
					g/=(n*n);
					b/=(n*n);
					a/=(n*n);
					*((UW *)cnewdata)=(UW)((r & m_rmask)|(g & m_gmask)|(b & m_bmask)|(a & m_amask));
					cnewdata+=2;
				}
			}
			delete [] m_data;
			m_data=newdata;
		}
		break;
	case 24:
		break;
	case 32:
		break;
	}
	m_w/=n;
	m_h/=n;
}

