//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
//
//	Comment :	LZSS comp & decom
//
//	Management of animations
//
// To do				:
//-----------------------------------------------------------------------------

#include <stdafx.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <g3lzss.h>

/**********************************************************************/

#define BLOCK_SIZE	32768


typedef struct
{
    size_t size;		// current read or write size
    size_t buf_size;	// utiliser pour la sortie
    int	bits;			// current value to read or write bits
    int bits_cnt;		// current bits count
    size_t bits_offs;	// where to put the bits value
    unsigned char *blk_add;	// start address of block ptr
    unsigned char *blk_ptr;	// ptr on the current byte
} BITS;


/* ************************************************
 *
 *       Input bits function
 *
 * ***********************************************/



inline int InputByte(BITS *ptr)
{
	int b;
    if (ptr->size == 0)
    {
        b=EOF;
    }
    else
    {
    	b = *(ptr->blk_ptr++);
		ptr->size--;
    }
    return b;
}


inline int InputWord(BITS *ptr)		// les 2 fonctions suivantes
{					// ne controle pas le EOF
	int w;
    w = InputByte(ptr);
    w += InputByte(ptr)<<8;
    return w;
}

inline size_t InputDword(BITS *ptr)
{
	size_t w;
    w  = InputByte(ptr);
    w += InputByte(ptr)<<8;
    w += InputByte(ptr)<<16;
    w += InputByte(ptr)<<24;
    return w;
}



inline size_t InitInBit(BITS *ptr, unsigned char *in, size_t size)
{
    ptr->size = size;	// sert de compteur pour savoir si a la fin
    ptr->buf_size = size;
    ptr->blk_add = in;
    ptr->blk_ptr = ptr->blk_add;
    ptr->bits_offs = 0;	// sans importance
    ptr->bits = 0;	// idem
    ptr->bits_cnt = 0;

    return (InputDword(ptr));	// renvoie la taille originale...
}


inline void InitInByte(BITS *ptr, unsigned char *in, size_t size)
{
    ptr->size = size;	// sert de compteur pour savoir si a la fin
    ptr->buf_size = size;
    ptr->blk_add = in;
    ptr->blk_ptr = ptr->blk_add;
    ptr->bits_offs = 0;	// sans importance
    ptr->bits = 0;	// idem
    ptr->bits_cnt = 0;
}


inline int InputBit(BITS *ptr)
{
	int b;
    if (ptr->bits_cnt == 0)
    {
		ptr->bits = InputByte(ptr);
		ptr->bits_cnt = 8;
    }
    ptr->bits_cnt--;
    b = ptr->bits & 1;
    ptr->bits >>= 1;
    return b;
}



/* ************************************************
 *
 *       Output bits function
 *
 * ***********************************************/



inline void OutputByte(BITS *ptr, int byte)
{

    if (ptr->size >= ptr->buf_size)
    {
        ptr->buf_size += BLOCK_SIZE;
		if (ptr->blk_add = (unsigned char*)realloc(ptr->blk_add, ptr->buf_size))
		{
			ptr->blk_ptr = ptr->blk_add + ptr->size ;
		}
		else
		{
			fprintf(stderr, "Can't allocate memory for Output...\n");
			exit(1);
		}
    }
    *(ptr->blk_ptr++) = byte;
    ptr->size++;
}


inline void OutputWord(BITS *ptr, int word)
{
    OutputByte(ptr, word);
    OutputByte(ptr, word>>8);
}


inline void OutputDword(BITS *ptr, size_t dword)
{
    OutputByte(ptr, dword);
    OutputByte(ptr, dword>>8);
    OutputByte(ptr, dword>>16);
    OutputByte(ptr, dword>>24);
}


inline void InitOutBit(BITS *ptr, size_t ori_size)
{
    ptr->size = 0;
    ptr->buf_size = 0;
    ptr->bits = 0;
    ptr->bits_cnt = 0;
    ptr->blk_add = NULL;

    OutputDword(ptr, ori_size);	// taille du fichier original

    ptr->bits_offs = ptr->size;
    OutputByte(ptr, 0);		// reserve la place les bits
}


inline void InitOutByte(BITS *ptr)
{
    ptr->size = 0;
    ptr->buf_size = 0;
    ptr->bits = 0;
    ptr->bits_cnt = 0;
    ptr->bits_offs = 0;
    ptr->blk_add = NULL;
}



inline void FlushOutBit(BITS *ptr)
{
    *(ptr->blk_add+ptr->bits_offs) = ptr->bits;
}



inline void OutputBit(BITS *ptr, int bit)
{
    if (ptr->bits_cnt == 8 )
    {
		FlushOutBit(ptr);
		ptr->bits_offs = ptr->size;
		OutputByte(ptr, 0);

		ptr->bits = 0;
		ptr->bits_cnt = 0;
    }

    ptr->bits += bit<<ptr->bits_cnt;
    ptr->bits_cnt++;

}


/*
 * Various constants used to define the compression parameters.  The
 * INDEX_BIT_COUNT tells how many bits we allocate to indices into the
 * text window.  This directly determines the WINDOW_SIZE.  The
 * LENGTH_BIT_COUNT tells how many bits we allocate for the length of
 * an encode phrase. This determines the size of the look ahead buffer.
 * The TREE_ROOT is a special node in the tree that always points to
 * the root node of the binary phrase tree.  END_OF_STREAM is a special
 * index used to flag the fact that the file has been completely
 * encoded, and there is no more data.  LZSS_UNUSED is the null index for
 * the tree. MOD_WINDOW() is a macro used to perform arithmetic on tree
 * indices.
 *
 */

#ifndef LZSS_INLINE
#define LZSS_INLINE			 inline
#endif

#ifndef INDEX_BIT_COUNT
#define INDEX_BIT_COUNT      10
#endif
#define LENGTH_BIT_COUNT     (16-INDEX_BIT_COUNT)
#define INDEX_BIT_MASK       ((1<<INDEX_BIT_COUNT)-1)
#define LENGTH_BIT_MASK      ((1<<LENGTH_BIT_COUNT)-1)
#define WINDOW_SIZE          ( 1 << INDEX_BIT_COUNT )
#define RAW_LOOK_AHEAD_SIZE  ( 1 << LENGTH_BIT_COUNT )
//#define BREAK_EVEN           ( ( 1 + INDEX_BIT_COUNT + LENGTH_BIT_COUNT ) / 9 )
#define BREAK_EVEN			2			// better result
#define LOOK_AHEAD_SIZE      ( RAW_LOOK_AHEAD_SIZE + BREAK_EVEN )
#define TREE_ROOT            WINDOW_SIZE
#define END_OF_STREAM        0
#define LZSS_UNUSED          0
#define MOD_WINDOW( a )      ( ( a ) & ( WINDOW_SIZE - 1 ) )


/*
 * These are the two global data structures used in this program.
 * The window[] array is exactly that, the window of previously seen
 * text, as well as the current look ahead text.  The tree[] structure
 * contains the binary tree of all of the strings in the window sorted
 * in order.
*/

//unsigned char window[ WINDOW_SIZE ];

typedef struct
{
int parent;
int smaller_child;
int larger_child;
}
TREE;

//} tree[ WINDOW_SIZE + 1 ];

/*
 * Function prototypes for ANSI C compilers
 */

void InitTree(TREE *tree,int r );
void ContractNode(TREE *tree,int old_node, int new_node );
void ReplaceNode(TREE *tree,int old_node, int new_node );
int FindNextNode(TREE *tree,int node );
void DeleteString(TREE *tree,int p );
int AddString(TREE *tree,unsigned char *window,int new_node, int *match_position );
void CompressFile( BITS *input, BITS *output );
void ExpandFile( BITS *input, BITS *output );


/*
 * Since the tree is static data, it comes up with every node
 * initialized to 0, which is good, since 0 is the LZSS_UNUSED code.
 * However, to make the tree really usable, a single phrase has to be
 * added to the tree so it has a root node.  That is done right here.
*/
void InitTree(TREE *tree,int r )
{
    memset(tree,0,sizeof(TREE)*WINDOW_SIZE + 1);
    tree[ TREE_ROOT ].larger_child = r;
    tree[ r ].parent = TREE_ROOT;
    tree[ r ].larger_child = LZSS_UNUSED;
    tree[ r ].smaller_child = LZSS_UNUSED;
}

/*
 * This routine is used when a node is being deleted.  The link to
 * its descendant is broken by pulling the descendant in to overlay
 * the existing link.
 */
void ContractNode(TREE *tree,int old_node,int new_node )
{
    tree[ new_node ].parent = tree[ old_node ].parent;
    if ( tree[ tree[ old_node ].parent ].larger_child == old_node )
        tree[ tree[ old_node ].parent ].larger_child = new_node;
    else
        tree[ tree[ old_node ].parent ].smaller_child = new_node;
    tree[ old_node ].parent = LZSS_UNUSED;
}

/*
 * This routine is also used when a node is being deleted.  However,
 * in this case, it is being replaced by a node that was not previously
 * in the tree.
 */
void ReplaceNode(TREE *tree,int old_node, int new_node )
{
    int parent;

    parent = tree[ old_node ].parent;
    if ( tree[ parent ].smaller_child == old_node )
        tree[ parent ].smaller_child = new_node;
    else
        tree[ parent ].larger_child = new_node;
    tree[ new_node ] = tree[ old_node ];
    tree[ tree[ new_node ].smaller_child ].parent = new_node;
    tree[ tree[ new_node ].larger_child ].parent = new_node;
    tree[ old_node ].parent = LZSS_UNUSED;
}

/*
 * This routine is used to find the next smallest node after the node
 * argument.  It assumes that the node has a smaller child.  We find
 * the next smallest child by going to the smaller_child node, then
 * going to the end of the larger_child descendant chain.
*/
int FindNextNode(TREE *tree,int node )
{
    int next;

    next = tree[ node ].smaller_child;
    while ( tree[ next ].larger_child != LZSS_UNUSED )
        next = tree[ next ].larger_child;
    return( next );
}

/*
 * This routine performs the classic binary tree deletion algorithm.
 * If the node to be deleted has a null link in either direction, we
 * just pull the non-null link up one to replace the existing link.
 * If both links exist, we instead delete the next link in order, which
 * is guaranteed to have a null link, then replace the node to be deleted
 * with the next link.
 */
void DeleteString(TREE *tree, int p )
{
    int  replacement;

    if ( tree[ p ].parent == LZSS_UNUSED )
        return;
    if ( tree[ p ].larger_child == LZSS_UNUSED )
        ContractNode(tree, p, tree[ p ].smaller_child );
    else if ( tree[ p ].smaller_child == LZSS_UNUSED )
        ContractNode(tree, p, tree[ p ].larger_child );
    else {
        replacement = FindNextNode(tree, p );
        DeleteString(tree, replacement );
        ReplaceNode(tree, p, replacement );
    }
}

/*
 * This where most of the work done by the encoder takes place.  This
 * routine is responsible for adding the new node to the binary tree.
 * It also has to find the best match among all the existing nodes in
 * the tree, and return that to the calling routine.  To make matters
 * even more complicated, if the new_node has a duplicate in the tree,
 * the old_node is deleted, for reasons of efficiency.
 */

int AddString(TREE *tree,unsigned char *window,int new_node, int *match_position )
{
    int i;
    int test_node;
    int delta;
    int match_length;
    int *child;

    if ( new_node == END_OF_STREAM )
        return( 0 );
    test_node = tree[ TREE_ROOT ].larger_child;
    match_length = 0;
    for ( ; ; ) {
        for ( i = 0 ; i < LOOK_AHEAD_SIZE ; i++ ) {
            delta = window[ MOD_WINDOW( new_node + i ) ] -
                    window[ MOD_WINDOW( test_node + i ) ];
            if ( delta != 0 )
                break;
        }
        if ( i >= match_length ) {
            match_length = i;
            *match_position = test_node;
            if ( match_length >= LOOK_AHEAD_SIZE ) {
                ReplaceNode(tree,test_node, new_node );
                return( match_length );
            }
        }
        if ( delta >= 0 )
            child = &tree[ test_node ].larger_child;
        else
            child = &tree[ test_node ].smaller_child;
        if ( *child == LZSS_UNUSED ) {
            *child = new_node;
            tree[ new_node ].parent = test_node;
            tree[ new_node ].larger_child = LZSS_UNUSED;
            tree[ new_node ].smaller_child = LZSS_UNUSED;
            return( match_length );
        }
        test_node = *child;
    }
}

/*
 * This is the compression routine.  It has to first load up the look
 * ahead buffer, then go into the main compression loop.  The main loop
 * decides whether to output a single character or an index/length
 * token that defines a phrase.  Once the character or phrase has been
 * sent out, another loop has to run.  The second loop reads in new
 * characters, deletes the strings that are overwritten by the new
 * character, then adds the strings that are created by the new
 * character.
 */

void CompressFile(BITS *input, BITS *output )
{
    int i;
    int c;
    int look_ahead_bytes;
    int current_position;
    int replace_count;
    int match_length;
    int match_position;
	TREE *tree;
	unsigned char *window;

	tree=(TREE *)malloc(sizeof(TREE)*(WINDOW_SIZE + 1));
	window=(unsigned char *)malloc(sizeof(unsigned char)*(WINDOW_SIZE));
    
	current_position = 1;

    for ( i = 0 ; i < LOOK_AHEAD_SIZE ; i++ ) {
        if ( ( c = InputByte( input ) ) == EOF )
            break;
        window[ current_position + i ] = (unsigned char) c;
    }
    look_ahead_bytes = i;
    InitTree(tree, current_position );
    match_length = 0;
    match_position = 0;
    while ( look_ahead_bytes > 0 ) {
        if ( match_length > look_ahead_bytes )
            match_length = look_ahead_bytes;
        if ( match_length <= BREAK_EVEN ) {
            replace_count = 1;
            OutputBit( output, 1 );
            OutputByte( output, window[ current_position ] );
        } else {
	#if 0
	    if (!(MOD_WINDOW(current_position - match_position)))
	    {
			printf("Warning bad MOD_WINDOW\n");
	    }
	#endif

	    OutputBit( output, 0 );
	    OutputWord( output, MOD_WINDOW(current_position - match_position) +
	                (( match_length - ( BREAK_EVEN + 1 ) )<<INDEX_BIT_COUNT));
            replace_count = match_length;
        }
        for ( i = 0 ; i < replace_count ; i++ ) {
            DeleteString(tree, MOD_WINDOW( current_position + LOOK_AHEAD_SIZE ) );
            if ( ( c = InputByte( input ) ) == EOF )
                look_ahead_bytes--;
            else
                window[ MOD_WINDOW( current_position + LOOK_AHEAD_SIZE ) ]
                        = (unsigned char) c;
            current_position = MOD_WINDOW( current_position + 1 );
            if ( look_ahead_bytes )
                match_length = AddString(tree,window,current_position, &match_position );
        }
    };
    OutputBit( output, 0 );
    OutputWord( output, END_OF_STREAM );

	free(window);
	free(tree);
}

/*
 * This is the expansion routine for the LZSS algorithm.  All it has
 * to do is read in flag bits, decide whether to read in a character or
 * a index/length pair, and take the appropriate action.
*/

void ExpandFile(BITS *input, BITS *output )
{
    int i;
    int current_position;
    int c;
    int temp;
    int match_length;
    int match_position;

	unsigned char *window;
	window=(unsigned char *)malloc(sizeof(unsigned char)*(WINDOW_SIZE));

    current_position = 1;
    for ( ; ; ) {
        if ( InputBit( input ) ) {
            c = (int) InputByte( input );
            OutputByte( output, c );
            window[ current_position ] = (unsigned char) c;
            current_position = MOD_WINDOW( current_position + 1 );
        } else {
	    temp = InputWord( input );
            match_position = (int) temp & INDEX_BIT_MASK;
            if ( match_position == END_OF_STREAM )
                break;
	    match_position = MOD_WINDOW(current_position - match_position);
	    match_length = (int) (temp>>INDEX_BIT_COUNT) & LENGTH_BIT_MASK;
            match_length += BREAK_EVEN;
            for ( i = 0 ; i <= match_length ; i++ ) {
                c = window[ MOD_WINDOW( match_position + i ) ];
                OutputByte( output, c );
                window[ current_position ] = (unsigned char) c;
                current_position = MOD_WINDOW( current_position + 1 );
            }
        }
    }
	free(window);
}


unsigned char *lzss_pack(unsigned char *srcbuffer,size_t srcsize)
{
	BITS input;
	BITS output;
	// Compress Data
	InitInByte(&input, (unsigned char*)srcbuffer,srcsize);
	InitOutBit(&output,srcsize);
	CompressFile(&input, &output);
	FlushOutBit(&output);
	output.blk_add = (unsigned char*)realloc(output.blk_add,output.size+8);
	memmove((unsigned char*)output.blk_add+8,output.blk_add,output.size);
	*((long *)output.blk_add)=LZSSCODE;
	*((long *)output.blk_add+1)=output.size+8;
	return output.blk_add;
}

size_t lzss_getpacksize(unsigned char *buffer)
{
	return *((unsigned int *)buffer+1);
}

int lzss_ispacked(unsigned char *buffer)
{
	return (*((unsigned int *)buffer)&0xFFFFFFF0)==(LZSSCODE&0xFFFFFFF0);
}

unsigned char *lzss_unpack(unsigned char *srcbuffer)
{
	BITS input;
	BITS output;
	unsigned int header=*((unsigned int *)srcbuffer);
	unsigned int len=*(((unsigned int *)srcbuffer)+1);
	if((header&0xFFFFFFF0)==(LZSSCODE&0xFFFFFFF0) && len>8)
	{
		InitInBit(&input, (unsigned char*)srcbuffer+8,len);
		InitOutByte(&output);
		ExpandFile(&input,&output);
		return output.blk_add;
	}
	return NULL;
}


