//-----------------------------------------------------------------------------
// (C) 1997	Levastre Claude
//
//
//	Comment : matrix33+matrix43+v3d+quaternion+rotation
//
// Last modified		:
// To do				:
//-----------------------------------------------------------------------------

#include <stdafx.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#include <g3mat.h>
#include <g3utils.h>

//--------------------------------------------------------------------------------
// CV3D  
//--------------------------------------------------------------------------------

Cv3d spline(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3,Cv3d const &v4,CsplineFactor const &sf)
{
	Cv3d res;
	res.x=v1.x*sf.getm1()+v2.x*sf.getm2()+v3.x*sf.getm3()+v4.x*sf.getm4();
	res.y=v1.y*sf.getm1()+v2.y*sf.getm2()+v3.y*sf.getm3()+v4.y*sf.getm4();
	res.z=v1.z*sf.getm1()+v2.z*sf.getm2()+v3.z*sf.getm3()+v4.z*sf.getm4();
	return(res);
}


Cv3d spline(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3,Cv3d const &v4,float t)
{
	float t2,t3;
	float m1,m2,m3,m4;
	Cv3d res;
	t2=t*t;
	t3=t2*t;
	m1=((1.0F*t2)-(0.5F*t3)-(0.5F*t));
	m2=((1.5F*t3)-(2.5F*t2)+1.0F);
	m3=((2.0F*t2)-(1.5F*t3)+(0.5F*t));
	m4=((0.5F*t3)-(0.5F*t2));

	res.x=v1.x*m1+v2.x*m2+v3.x*m3+v4.x*m4;
	res.y=v1.y*m1+v2.y*m2+v3.y*m3+v4.y*m4;
	res.z=v1.z*m1+v2.z*m2+v3.z*m3+v4.z*m4;

	return(res);
}

int Cv3d::fread(G3FILE *handle)
{
	UW version;
	g3fread(&version,sizeof(version),1,handle);
	switch(version)
	{
	case 1:
		{
		int count=0;
		count+=g3fread(&x,sizeof(x),1,handle);
		count+=g3fread(&y,sizeof(y),1,handle);
		count+=g3fread(&z,sizeof(z),1,handle);
		if(count==3)
			return(1);
		}
		break;
	default:
		::MessageBox(NULL,"V3D : UNKNOWN VERSION","ERROR",MB_ICONWARNING|MB_OK);
		exit(0);
		break;

	}
	return(0);
}

int Cv3d::fwrite(G3FILE *handle)
{
	UW version=1;
	g3fwrite(&version,sizeof(version),1,handle);
	int count=0;
	count+=g3fwrite(&x,sizeof(x),1,handle);
	count+=g3fwrite(&y,sizeof(y),1,handle);
	count+=g3fwrite(&z,sizeof(z),1,handle);
	if(count==3)
		return(1);
	return(0);
}

void Cv3d::normalize(void)
{
	float length = (float)(sqrt(x*x+y*y+z*z));
	if (length!=0.0F)
	{
		length = 1.0F/length;
		x*=length;
		y*=length;
		z*=length;
	}
}

void Cv3d::amortize(Cv3d const &value,float ratio,float time)
{
	float r=ratio*time;
	x=floatamortize(x,value.x,r);
	y=floatamortize(y,value.y,r);
	z=floatamortize(z,value.z,r);
}

void Cv3d::adjustto(Cv3d const &wantedvalue,Cv3d const &step,float time)
{
	x=floatadjustto(x,wantedvalue.x,step.x,time);
	y=floatadjustto(y,wantedvalue.y,step.y,time);
	z=floatadjustto(z,wantedvalue.z,step.z,time);
}

void Cv3d::setinabsrange(Cv3d const &max)
{
	x=floatsetinabsrange(x,max.x);
	y=floatsetinabsrange(y,max.y);
	z=floatsetinabsrange(z,max.z);
}

void Cv3d::setinrange(Cv3d const &min,Cv3d const &max)
{
	x=floatsetinrange(x,min.x,max.x);
	y=floatsetinrange(y,min.y,max.y);
	z=floatsetinrange(z,min.z,max.z);
}

void Cv3d::rotatex(float angle)
{
	//  X=x;
	//  Y=y*cosx-z*sinx;
	//  Z=y*sinx+z*cosx;
	float s;
	float c;
	float tmp;
	s=(float)sin(angle);
	c=(float)cos(angle);
	tmp=y;
	y=y*c-z*s;
	z=tmp*s+z*c;
}

void Cv3d::rotatey(float angle)
{
	//  X=x*cosy+z*siny
	//  Y=y;
	//  Z=-x*siny+z*cosy
	float s;
	float c;
	float tmp;
	s=(float)sin(angle);
	c=(float)cos(angle);
	tmp=x;
	x=x*c+z*s;
	z=-tmp*s+z*c;
}

void Cv3d::rotatez(float angle)
{
	//  X=x*cosz-y*sinz;
	//  Y=x*sinz+y*cosz;
	//  Z=z;
	float s;
	float c;
	float tmp;
	s=(float)sin(angle);
	c=(float)cos(angle);
	tmp=x;
	x=x*c-y*s;
	y=tmp*s+y*c;
}

//--------------------------------------------------------------------------------
// CMAT33
//--------------------------------------------------------------------------------


void Cmat33::set(float _value)
{
	m11 = _value;m12 = 0.0F;m13 = 0.0F;
	m21 = 0.0F;m22 = _value;m23 = 0.0F;
	m31 = 0.0F;m32 = 0.0F;m33 = _value;
}

Cmat33::Cmat33(float _value)
{
	set(_value);
}


void Cmat33::set(Cquat const &_quater)
{
	{
		float xx=_quater.x*_quater.x;
		float zz=_quater.z*_quater.z;
		float yy=_quater.y*_quater.y;
		xx+=xx;
		yy+=yy;
		zz+=zz;
		m11=1.0F-(yy+zz);
		m22=1.0F-(xx+zz);
		m33=1.0F-(xx+yy);
	}

	{
		float xy=_quater.x*_quater.y;
		float zw=_quater.z*_quater.w;
		xy+=xy;
		zw+=zw;
		m12=(xy-zw);
		m21=(xy+zw);
	}

	{
		float yz=_quater.y*_quater.z;
		float xw=_quater.x*_quater.w;
		yz+=yz;
		xw+=xw;
		m23=(yz-xw);
		m32=(yz+xw);
	}

	{
		float xz=_quater.x*_quater.z;
		float yw=_quater.y*_quater.w;
		xz+=xz;
		yw+=yw;
		m13=(xz+yw);
		m31=(xz-yw);
	}
}

Cmat33::Cmat33(Cquat const &_quater)
{
	set(_quater);
}

void Cmat33::setrx(float angle)
{
	//  X=x;
	//  Y=y*cosx-z*sinx;
	//  Z=y*sinx+z*cosx;
	float s;
	float c;
	s=(float)sin(angle);
	c=(float)cos(angle);
	m11=1.0F;
	m12=0.0F;
	m13=0.0F;
	m21=0.0F;
	m22=c;
	m23=s;
	m31=0.0F;
	m32=-s;
	m33=c;
}

void Cmat33::setry(float angle)
{
	//  X=x*cosy+z*siny
	//  Y=y;
	//  Z=-x*siny+z*cosy
	float s;
	float c;
	s=(float)sin(angle);
	c=(float)cos(angle);
	m11=c;
	m12=0.0F;
	m13=-s;
	m21=0.0F;
	m22=1.0F;
	m23=0.0F;
	m31=s;
	m32=0.0F;
	m33=c;
}

void Cmat33::setrz(float angle)
{
	//  X=x*cosz-y*sinz;
	//  Y=x*sinz+y*cosz;
	//  Z=z;
	float s;
	float c;
	s=(float)sin(angle);
	c=(float)cos(angle);
	m11=c;
	m12=s;
	m13=0.0F;
	m21=-s;
	m22=c;
	m23=0.0F;
	m31=0.0F;
	m32=0.0F;
	m33=1.0F;
}

Cmat33::Cmat33(Crot const &_rot)      // Z X Y
{
	set(_rot);
}

void Cmat33::set(Crot const &_rot)      // Z X Y
{
  float sinx,siny,sinz;
  float cosx,cosy,cosz;
  float tmp;
  sinx=(float)sin(_rot.x);
  cosx=(float)cos(_rot.x);
  siny=(float)sin(_rot.y);
  cosy=(float)cos(_rot.y);
  sinz=(float)sin(_rot.z);
  cosz=(float)cos(_rot.z);
  tmp=sinx*siny;
  m11=cosz*cosy+sinz*tmp;
  m21=cosz*tmp-sinz*cosy;
  m31=cosx*siny;
  m12=sinz*cosx;
  m22=cosz*cosx;
  m32=-sinx;
  tmp=sinx*cosy;
  m13=tmp*sinz-cosz*siny;
  m23=sinz*siny+tmp*cosz;
  m33=cosx*cosy;
}

Cmat33 &Cmat33::operator += (Cmat33 const &_mat)
{
	m11 += _mat.m11;
	m12 += _mat.m12;
	m13 += _mat.m13;
	m21 += _mat.m21;
	m22 += _mat.m22;
	m23 += _mat.m23;
	m31 += _mat.m31;
	m32 += _mat.m32;
	m33 += _mat.m33;
	return (*this);
}

Cmat33 &Cmat33::operator -= (Cmat33 const &_mat)
{
	m11 -= _mat.m11;
	m12 -= _mat.m12;
	m13 -= _mat.m13;
	m21 -= _mat.m21;
	m22 -= _mat.m22;
	m23 -= _mat.m23;
	m31 -= _mat.m31;
	m32 -= _mat.m32;
	m33 -= _mat.m33;
	return (*this);
}

Cmat33 &Cmat33::operator *= (float _k)
{
	m11 *= _k;
	m12 *= _k;
	m13 *= _k;
	m21 *= _k;
	m22 *= _k;
	m23 *= _k;
	m31 *= _k;
	m32 *= _k;
	m33 *= _k;
	return (*this);
}

Cmat33 Cmat33::inverse(void)const
{
	Cmat33   res;
	res.m11 = m11;
	res.m12 = m21;
	res.m13 = m31;
	res.m21 = m12;
	res.m22 = m22;
	res.m23 = m32;
	res.m31 = m13;
	res.m32 = m23;
	res.m33 = m33;
	return (res);
}

void Cmat33::xyzoom(float _xzoom,float _yzoom)
{
	m11*=_xzoom;
	m21*=_xzoom;
	m31*=_xzoom;
	m12*=_yzoom;
	m22*=_yzoom;
	m32*=_yzoom;
}

void Cmat33::xyzzoom(float _xzoom,float _yzoom,float _zzoom)
{
	m11*=_xzoom;
	m21*=_xzoom;
	m31*=_xzoom;
	m12*=_yzoom;
	m22*=_yzoom;
	m32*=_yzoom;
	m13*=_zzoom;
	m23*=_zzoom;
	m33*=_zzoom;
}

void Cmat33::set(Cv3d const &a,Cv3d const &b,Cv3d const &c)
{
	Cv3d v1,n1,n2,n3;
	// compute normal of side
	v1=b-a;
	n1=v1 ^ (c-a);
	n1.normalize();
	// normalize vector A->B
	n2=v1;
	n2.normalize();
	// find normal n1,n2
	n3=n1 ^ n2;
	n3.normalize();

	m13=n1.x;
	m23=n1.y;
	m33=n1.z;
	m11=n2.x;
	m21=n2.y;
	m31=n2.z;
	m12=n3.x;
	m22=n3.y;
	m32=n3.z;
}

Cmat33::Cmat33(Cv3d const &a,Cv3d const &b,Cv3d const &c)
{
	set(a,b,c);
}


Cmat33 operator - (Cmat33 const &_mat1, Cmat33 const &_mat2)
{
	Cmat33   res;
	res.m11 = _mat1.m11 - _mat2.m11;
	res.m12 = _mat1.m12 - _mat2.m12;
	res.m13 = _mat1.m13 - _mat2.m13;
	res.m21 = _mat1.m21 - _mat2.m21;
	res.m22 = _mat1.m22 - _mat2.m22;
	res.m23 = _mat1.m23 - _mat2.m23;
	res.m31 = _mat1.m31 - _mat2.m31;
	res.m32 = _mat1.m32 - _mat2.m32;
	res.m33 = _mat1.m33 - _mat2.m33;
	return (res);
}

Cmat33 operator + (Cmat33 const &_mat1, Cmat33 const &_mat2)
{
	Cmat33   res;
	res.m11 = _mat1.m11 + _mat2.m11;
	res.m12 = _mat1.m12 + _mat2.m12;
	res.m13 = _mat1.m13 + _mat2.m13;
	res.m21 = _mat1.m21 + _mat2.m21;
	res.m22 = _mat1.m22 + _mat2.m22;
	res.m23 = _mat1.m23 + _mat2.m23;
	res.m31 = _mat1.m31 + _mat2.m31;
	res.m32 = _mat1.m32 + _mat2.m32;
	res.m33 = _mat1.m33 + _mat2.m33;
	return (res);
}

Cmat33 operator * (Cmat33 const &mat1, Cmat33 const &mat2)
{
	Cmat33   res;

	#if 1
			_asm
			{
			push	ebx
			push	esi
			push	edi
			mov		esi,[mat1]
			mov		ebx,[mat2]
			lea		edi,[res]

									//  0   1   2   3   4   5   6   7
			fld		[esi]Cmat.m11
			fmul	[ebx]Cmat.m11	// M11
			fld		[esi]Cmat.m11
			fmul	[ebx]Cmat.m12	// M12 M11
			fld		[esi]Cmat.m11
			fmul	[ebx]Cmat.m13	// M13 M12 M11

			fld		[esi]Cmat.m12
			fmul	[ebx]Cmat.m21	//M21 M13 M12 M11
			fld		[esi]Cmat.m12
			fmul	[ebx]Cmat.m22	// M22 M21 M13 M12 M11
			fld		[esi]Cmat.m12
			fmul	[ebx]Cmat.m23	// M23 M22 M21 M13 M12 M11

			fxch	st(2)			// M21 M22 M23 M13 M12 M11
			faddp	st(5), st		// M22 M23 M13 M12 s1
			faddp	st(3), st		// M23 M13 s2  s1
			faddp	st(1), st		// s3  s2  s1
			fld		[esi]Cmat.m13
			fmul	[ebx]Cmat.m31	//31 s3  s2  s1
			fld		[esi]Cmat.m13
			fmul	[ebx]Cmat.m32	//M32 M31 s3  s2  s1
			fld		[esi]Cmat.m13
			fmul	[ebx]Cmat.m33	// M33 M32 M31 s3  s2  s1
			fxch	st(2)			// M31 M32 M33 s3  s2  s1
			faddp	st(5), st		// M32 M33 s3  s2  S1
			faddp	st(3), st		// M33 s3  S2  S1
			faddp	st(1), st		// S3  S2  S1
			fxch	st(2)			// S1  S2  S3
			fstp	[edi]Cmat.m11
			fstp	[edi]Cmat.m12
			fstp	[edi]Cmat.m13
						//  0   1   2   3   4   5   6   7
			fld		[esi]Cmat.m21
			fmul	[ebx]Cmat.m11	// M11
			fld		[esi]Cmat.m21
			fmul	[ebx]Cmat.m12	// M12 M11
			fld		[esi]Cmat.m21
			fmul	[ebx]Cmat.m13	// M13 M12 M11

			fld		[esi]Cmat.m22
			fmul	[ebx]Cmat.m21	// M21 M13 M12 M11
			fld		[esi]Cmat.m22
			fmul	[ebx]Cmat.m22	// M22 M21 M13 M12 M11
			fld		[esi]Cmat.m22
			fmul	[ebx]Cmat.m23	// M23 M22 M21 M13 M12 M11
			fxch	st(2)			// M21 M22 M23 M13 M12 M11
			faddp	st(5), st		// M22 M23 M13 M12 s1
			faddp	st(3), st		// M23 M13 s2  s1
			faddp	st(1), st		// s3  s2  s1

			fld		[esi]Cmat.m23
			fmul	[ebx]Cmat.m31	// M31 s3  s2  s1
			fld		[esi]Cmat.m23
			fmul	[ebx]Cmat.m32	// M32 M31 s3  s2  s1
			fld		[esi]Cmat.m23
			fmul	[ebx]Cmat.m33	// M33 M32 M31 s3  s2  s1
			fxch	st(2)		// M31 M32 M33 s3  s2  s1
			faddp	st(5), st	// M32 M33 s3  s2  S1
			faddp	st(3), st	// M33 s3  S2  S1
			faddp	st(1), st	// S3  S2  S1
			fxch	st(2)		// S1  S2  S3

			fstp	[edi]Cmat.m21
			fstp	[edi]Cmat.m22
			fstp	[edi]Cmat.m23
								//  0   1   2   3   4   5   6   7
			fld		[esi]Cmat.m31
			fmul	[ebx]Cmat.m11	// M11
			fld		[esi]Cmat.m31
			fmul	[ebx]Cmat.m12	// M12 M11
			fld		[esi]Cmat.m31
			fmul	[ebx]Cmat.m13	// M13 M12 M11

			fld		[esi]Cmat.m32
			fmul	[ebx]Cmat.m21	// M21 M13 M12 M11
			fld		[esi]Cmat.m32
			fmul	[ebx]Cmat.m22	// M22 M21 M13 M12 M11
			fld		[esi]Cmat.m32
			fmul	[ebx]Cmat.m23	// M23 M22 M21 M13 M12 M11
			fxch	st(2)		// M21 M22 M23 M13 M12 M11
			faddp	st(5), st	// M22 M23 M13 M12 s1
			faddp	st(3), st	// M23 M13 s2  s1
			faddp	st(1), st	// s3  s2  s1

			fld		[esi]Cmat.m33
			fmul	[ebx]Cmat.m31	// M31 s3  s2  s1
			fld		[esi]Cmat.m33
			fmul	[ebx]Cmat.m32	// M32 M31 s3  s2  s1
			fld		[esi]Cmat.m33
			fmul	[ebx]Cmat.m33	// M33 M32 M31 s3  s2  s1
			fxch	st(2)		// M31 M32 M33 s3  s2  s1
			faddp	st(5), st	// M32 M33 s3  s2  S1
			faddp	st(3), st	// M33 s3  S2  S1
			faddp	st(1), st	// S3  S2  S1
			fxch	st(2)		// S1  S2  S3

			fstp	[edi]Cmat.m31
			fstp	[edi]Cmat.m32
			fstp	[edi]Cmat.m33

			pop		edi
			pop		esi
			pop		ebx
		
			}
	#else
		res.m11 = ()mat1.m11*mat2.m11 + mat1.m12*mat2.m21 + mat1.m13*mat2.m31;
		res.m12 = mat1.m11*mat2.m12 + mat1.m12*mat2.m22 + mat1.m13*mat2.m32;
		res.m13 = mat1.m11*mat2.m13 + mat1.m12*mat2.m23 + mat1.m13*mat2.m33;

		res.m21 = mat1.m21*mat2.m11 + mat1.m22*mat2.m21 + mat1.m23*mat2.m31;
		res.m22 = mat1.m21*mat2.m12 + mat1.m22*mat2.m22 + mat1.m23*mat2.m32;
		res.m23 = mat1.m21*mat2.m13 + mat1.m22*mat2.m23 + mat1.m23*mat2.m33;

		res.m31 = mat1.m31*mat2.m11 + mat1.m32*mat2.m21 + mat1.m33*mat2.m31;
		res.m32 = mat1.m31*mat2.m12 + mat1.m32*mat2.m22 + mat1.m33*mat2.m32;
		res.m33 = mat1.m31*mat2.m13 + mat1.m32*mat2.m23 + mat1.m33*mat2.m33;
	#endif
	return (res);
}


Cmat33 operator * (Cmat33 const &mat1, float k)
{
	Cmat33 res;
	res.m11 = mat1.m11*k;
	res.m12 = mat1.m12*k;
	res.m13 = mat1.m13*k;

	res.m21 = mat1.m21*k;
	res.m22 = mat1.m22*k;
	res.m23 = mat1.m23*k;

	res.m31 = mat1.m31*k;
	res.m32 = mat1.m32*k;
	res.m33 = mat1.m33*k;
	return res;
}


Cv3d operator * (Cv3d const &v,Cmat33 const &mat)
{
	Cv3d res;
	res.x = v.x*mat.m11 + v.y*mat.m21 + v.z*mat.m31;
	res.y = v.x*mat.m12 + v.y*mat.m22 + v.z*mat.m32;
	res.z = v.x*mat.m13 + v.y*mat.m23 + v.z*mat.m33;
	return res;
}

int Cmat33::fread(G3FILE *handle)
{
	UW version;
	g3fread(&version,sizeof(version),1,handle);
	switch(version)
	{
	case 1:
		{
		int count=0;
		count+=g3fread(&m11,sizeof(m11),1,handle);
		count+=g3fread(&m12,sizeof(m12),1,handle);
		count+=g3fread(&m13,sizeof(m13),1,handle);
		count+=g3fread(&m21,sizeof(m21),1,handle);
		count+=g3fread(&m22,sizeof(m22),1,handle);
		count+=g3fread(&m23,sizeof(m23),1,handle);
		count+=g3fread(&m31,sizeof(m31),1,handle);
		count+=g3fread(&m32,sizeof(m32),1,handle);
		count+=g3fread(&m33,sizeof(m33),1,handle);
		if(count==9)
			return(1);
		}
		break;
	default:
		::MessageBox(NULL,"MAT : UNKNOWN VERSION","ERROR",MB_ICONWARNING|MB_OK);
		exit(0);
		break;
	}
	return(0);
}

int Cmat33::fwrite(G3FILE *handle)
{
	UW version=1;
	g3fwrite(&version,sizeof(version),1,handle);
	int count=0;
	count+=g3fwrite(&m11,sizeof(m11),1,handle);
	count+=g3fwrite(&m12,sizeof(m12),1,handle);
	count+=g3fwrite(&m13,sizeof(m13),1,handle);
	count+=g3fwrite(&m21,sizeof(m21),1,handle);
	count+=g3fwrite(&m22,sizeof(m22),1,handle);
	count+=g3fwrite(&m23,sizeof(m23),1,handle);
	count+=g3fwrite(&m31,sizeof(m31),1,handle);
	count+=g3fwrite(&m32,sizeof(m32),1,handle);
	count+=g3fwrite(&m33,sizeof(m33),1,handle);
	if(count==9)
		return(1);
	return(0);
}


//--------------------------------------------------------------------------------
// CMAT
//--------------------------------------------------------------------------------

Cmat::Cmat(Cv3d const &_trs)
{
	mat33.set(1.0F);
	trs = _trs;
}
void Cmat::set(Cv3d const &_trs)
{
	mat33.set(1.0F);
	trs = _trs;
}

Cmat::Cmat(float _value)
{
	mat33.set(_value);
	trs.set(0.0F);
}

void Cmat::set(float _value)
{
	mat33.set(_value);
	trs.set(0.0F);
}

void Cmat::set()
{
	mat33.set(1.0F);
	trs.set(0.0F);
}


void Cmat::set(Cquat const &_quater)
{
	mat33.set(_quater);
	trs.set(0.0F);
}

Cmat::Cmat(Cquat const &_quater)
{
	set(_quater);
}

Cmat::Cmat(Cquat const &_quater,Cv3d const &_trs)
{
	mat33.set(_quater);
	trs=_trs;
}

void Cmat::set(Cquat const &_quater,Cv3d const &_trs)
{
	mat33.set(_quater);
	trs=_trs;
}

Cmat::Cmat(Cv3d const &_trs1,Cquat const &_quater,Cv3d const &_trs)
{
	mat33.set(_quater);
	trs.x=_trs.x+(m11*_trs1.x+m21*_trs1.y+m31*_trs1.z);
	trs.y=_trs.y+(m12*_trs1.x+m22*_trs1.y+m32*_trs1.z);
	trs.z=_trs.z+(m13*_trs1.x+m23*_trs1.y+m33*_trs1.z);
}

void Cmat::set(Cv3d const &_trs1,Cquat const &_quater,Cv3d const &_trs)
{
	mat33.set(_quater);
	trs.x=_trs.x+(m11*_trs1.x+m21*_trs1.y+m31*_trs1.z);
	trs.y=_trs.y+(m12*_trs1.x+m22*_trs1.y+m32*_trs1.z);
	trs.z=_trs.z+(m13*_trs1.x+m23*_trs1.y+m33*_trs1.z);
}

Cmat::Cmat(Cv3d const &_trs1,Cmat33 const &_mat33,Cv3d const &_trs)
{
	set(_trs1,_mat33,_trs);
}

void Cmat::set(Cv3d const &_trs1,Cmat33 const &_mat33,Cv3d const &_trs)
{
	mat33=_mat33;
	trs.x=_trs.x+(m11*_trs1.x+m21*_trs1.y+m31*_trs1.z);
	trs.y=_trs.y+(m12*_trs1.x+m22*_trs1.y+m32*_trs1.z);
	trs.z=_trs.z+(m13*_trs1.x+m23*_trs1.y+m33*_trs1.z);
}



Cmat::Cmat(Crot const &_rot,Cv3d const &_trs)
{
	set(_rot,_trs);
}

void Cmat::setrx(float angle)
{
	//  X=x;
	//  Y=y*cosx-z*sinx;
	//  Z=y*sinx+z*cosx;
/*	float sin;
	float cos;
	sincos.sincos(angle,&sin,&cos);
	m11=1.0F;
	m12=0.0F;
	m13=0.0F;
	m21=0.0F;
	m22=cos;
	m23=sin;
	m31=0.0F;
	m32=-sin;
	m33=cos;
*/
	mat33.setrx(angle);
	trs.set(0.0F);
}

void Cmat::setry(float angle)
{
	//  X=x*cosy+z*siny
	//  Y=y;
	//  Z=-x*siny+z*cosy
	mat33.setry(angle);
	trs.set(0.0F);
}

void Cmat::setrz(float angle)
{
	//  X=x*cosz-y*sinz;
	//  Y=x*sinz+y*cosz;
	//  Z=z;
	mat33.setrz(angle);
	trs.set(0.0F);
}

void Cmat::set(Crot const &_rot,Cv3d const &_trs)
{
  mat33.set(_rot);
  trs=_trs;
}

Cmat::Cmat(Cv3d const &_trs1,Crot const &_rot,Cv3d const &_trs)
{
	set(_trs1,_rot,_trs);
}

void Cmat::set(Cv3d const &_trs1,Crot const &_rot,Cv3d const &_trs)
{
  mat33.set(_rot);
  trs.x=_trs.x+(m11*_trs1.x+m21*_trs1.y+m31*_trs1.z);
  trs.y=_trs.y+(m12*_trs1.x+m22*_trs1.y+m32*_trs1.z);
  trs.z=_trs.z+(m13*_trs1.x+m23*_trs1.y+m33*_trs1.z);
}



Cmat::Cmat(Crot const &_rot)      // Z X Y
{
	set(_rot);
}

void Cmat::set(Crot const &_rot)      // Z X Y
{
  mat33.set(_rot);
  trs.set(0.0F);
}

Cmat &Cmat::operator += (Cmat const &_mat)
{
	mat33 += _mat.mat33;
	trs += _mat.trs;
	return (*this);
}

Cmat &Cmat::operator -= (Cmat const &_mat)
{
	mat33 -= _mat.mat33;
	trs -= _mat.trs;
	return (*this);
}

Cmat &Cmat::operator *= (float _k)
{
	mat33 *= _k;
	trs *= _k;
	return (*this);
}

Cmat Cmat::inverse(void)const
{
	Cmat   res;
	res.m11 = m11;
	res.m12 = m21;
	res.m13 = m31;
	res.m21 = m12;
	res.m22 = m22;
	res.m23 = m32;
	res.m31 = m13;
	res.m32 = m23;
	res.m33 = m33;
	res.trs.x = -(m11*trs.x + m12*trs.y + m13*trs.z);
	res.trs.y = -(m21*trs.x + m22*trs.y + m23*trs.z);
	res.trs.z = -(m31*trs.x + m32*trs.y + m33*trs.z);
	return (res);
}

void Cmat::xyzoom(float _xzoom,float _yzoom)
{
	mat33.xyzoom(_xzoom,_yzoom);
	trs.x*=_xzoom;
	trs.y*=_yzoom;
}

void Cmat::xyzzoom(float _xzoom,float _yzoom,float _zzoom)
{
	mat33.xyzzoom(_xzoom,_yzoom,_zzoom);
	trs.xyzzoom(_xzoom,_yzoom,_zzoom);
}

void Cmat::set(Cv3d const &a,Cv3d const &b,Cv3d const &c)
{
	mat33.set(a,b,c);
	trs.set(0.0F);
}

Cmat::Cmat(Cv3d const &a,Cv3d const &b,Cv3d const &c)
{
	set(a,b,c);
}


Cmat operator - (Cmat const &_mat1, Cmat const &_mat2)
{
	Cmat   res;
	res.mat33 = _mat1.mat33 - _mat2.mat33;
	res.trs = _mat1.trs - _mat2.trs;
	return (res);
}

Cmat operator + (Cmat const &_mat1, Cmat const &_mat2)
{
	Cmat   res;
	res.mat33 = _mat1.mat33 + _mat2.mat33;
	res.trs = _mat1.trs + _mat2.trs;
	return (res);
}

Cmat operator * (Cmat const &mat1, Cmat const &mat2)
{
	Cmat   res;
	res.mat33 = mat1.mat33 * mat2.mat33;
	res.trs = mat1.trs*mat2;
	return (res);
}

Cmat operator * (Cmat const &mat1, Cmat33 const &mat2)
{
	Cmat   res;
	res.mat33 = mat1.mat33 * mat2;
	res.trs = mat1.trs*mat2;
	return (res);
}


Cmat operator * (Cmat const &mat, float k)
{
	Cmat   res;
	res.mat33 = mat.mat33 * k;
	res.trs = mat.trs * k;
	return (res);
}


int Cmat::fread(G3FILE *handle)
{
	UW version;
	g3fread(&version,sizeof(version),1,handle);
	switch(version)
	{
	case 1:
		{
		int count=0;
		count+=g3fread(&m11,sizeof(m11),1,handle);
		count+=g3fread(&m12,sizeof(m12),1,handle);
		count+=g3fread(&m13,sizeof(m13),1,handle);
		count+=g3fread(&m21,sizeof(m21),1,handle);
		count+=g3fread(&m22,sizeof(m22),1,handle);
		count+=g3fread(&m23,sizeof(m23),1,handle);
		count+=g3fread(&m31,sizeof(m31),1,handle);
		count+=g3fread(&m32,sizeof(m32),1,handle);
		count+=g3fread(&m33,sizeof(m33),1,handle);
		count+=g3fread(&trs.x,sizeof(trs.x),1,handle);
		count+=g3fread(&trs.y,sizeof(trs.y),1,handle);
		count+=g3fread(&trs.z,sizeof(trs.z),1,handle);
		if(count==12)
			return(1);
		}
		break;
	case 2:
		{
		int count=0;
		count+=g3fread(&mat33,sizeof(mat33),1,handle);
		count+=g3fread(&trs,sizeof(trs),1,handle);
		if(count==2)
			return(1);
		}
		break;
	default:
		::MessageBox(NULL,"MAT : UNKNOWN VERSION","ERROR",MB_ICONWARNING|MB_OK);
		exit(0);
		break;
	}
	return(0);
}

int Cmat::fwrite(G3FILE *handle)
{
	UW version=2;
	g3fwrite(&version,sizeof(version),1,handle);
	int count=0;
	count+=g3fwrite(&mat33,sizeof(mat33),1,handle);
	count+=g3fwrite(&trs,sizeof(trs),1,handle);
	if(count==2)
		return(1);
	return(0);
}

Cv3d max3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3)
{
	Cv3d res;
	res.x = MAX3(v1.x,v2.x,v3.x);
	res.y = MAX3(v1.y,v2.y,v3.y);
	res.z = MAX3(v1.z,v2.z,v3.z);
	return(res);
}

Cv3d min3(Cv3d const &v1,Cv3d const &v2,Cv3d const &v3)
{
	Cv3d res;
	res.x = MIN3(v1.x,v2.x,v3.x);
	res.y = MIN3(v1.y,v2.y,v3.y);
	res.z = MIN3(v1.z,v2.z,v3.z);
	return(res);
}

Cv3d max2(Cv3d const &v1,Cv3d const &v2)
{
	Cv3d res;
	res.x = MAX(v1.x,v2.x);
	res.y = MAX(v1.y,v2.y);
	res.z = MAX(v1.z,v2.z);
	return(res);
}


Cv3d min2(Cv3d const &v1,Cv3d const &v2)
{
	Cv3d res;
	res.x = MIN(v1.x,v2.x);
	res.y = MIN(v1.y,v2.y);
	res.z = MIN(v1.z,v2.z);
	return(res);
}


Cv3d normal3(Cv3d const &va,Cv3d const &vb,Cv3d const &vc)
{
		Cv3d res;
        res=(vb-va) ^ (vc-va);
		res.normalize();
		return(res);
}

//--------------------------------------------------------------------------------
// CQUATERNION
//--------------------------------------------------------------------------------
void Cquat::set(Cmat33 const &mat)
{
	float tr;
	tr=mat.m11+mat.m22+mat.m33;
	if(tr>=0.0F)
	{
		float s=(float)sqrt(tr+1);
		w=s*0.5F;
		s=0.5F/s;
		x=(mat.m32-mat.m23)*s;
		y=(mat.m13-mat.m31)*s;
		z=(mat.m21-mat.m12)*s;
	}
	else
	{
		float temp;
		int i=0;
		temp=mat.m11;
		if(mat.m22>mat.m11)
		{
			temp=mat.m22;
			i=1;
		}
		if(mat.m33>temp)
		{
			// m33 is bigger one
			float s=(float)sqrt(mat.m33-mat.m11-mat.m22+1.0F);
			z=s*0.5F;
			s=0.5F/s;
			x=(mat.m31+mat.m13)*s;
			y=(mat.m32+mat.m23)*s;
			w=(mat.m21-mat.m12)*s;
			return;
		}
		if(i)
		{
			// M22 is bigger one
			float s=(float)sqrt(mat.m22-mat.m33-mat.m11+1.0F);
			y=s*0.5F;
			s=0.5F/s;
			z=(mat.m23+mat.m32)*s;
			x=(mat.m21+mat.m12)*s;
			w=(mat.m13-mat.m31)*s;
			return;
		}
		else
		{
			// M11 is bigger one
			float s=(float)sqrt(mat.m11-mat.m22-mat.m33+1.0F);
			x=s*0.5F;
			s=0.5F/s;
			y=(mat.m12+mat.m21)*s;
			z=(mat.m31+mat.m13)*s;
			w=(mat.m32-mat.m23)*s;
			return;
		}
	}
}

void Cquat::set(Cmat const &mat)
{
	float tr;
	tr=mat.m11+mat.m22+mat.m33;
	if(tr>=0.0F)
	{
		float s=(float)sqrt(tr+1);
		w=s*0.5F;
		s=0.5F/s;
		x=(mat.m32-mat.m23)*s;
		y=(mat.m13-mat.m31)*s;
		z=(mat.m21-mat.m12)*s;
	}
	else
	{
		float temp;
		int i=0;
		temp=mat.m11;
		if(mat.m22>mat.m11)
		{
			temp=mat.m22;
			i=1;
		}
		if(mat.m33>temp)
		{
			// m33 is bigger one
			float s=(float)sqrt(mat.m33-mat.m11-mat.m22+1.0F);
			z=s*0.5F;
			s=0.5F/s;
			x=(mat.m31+mat.m13)*s;
			y=(mat.m32+mat.m23)*s;
			w=(mat.m21-mat.m12)*s;
			return;
		}
		if(i)
		{
			// M22 is bigger one
			float s=(float)sqrt(mat.m22-mat.m33-mat.m11+1.0F);
			y=s*0.5F;
			s=0.5F/s;
			z=(mat.m23+mat.m32)*s;
			x=(mat.m21+mat.m12)*s;
			w=(mat.m13-mat.m31)*s;
			return;
		}
		else
		{
			// M11 is bigger one
			float s=(float)sqrt(mat.m11-mat.m22-mat.m33+1.0F);
			x=s*0.5F;
			s=0.5F/s;
			y=(mat.m12+mat.m21)*s;
			z=(mat.m31+mat.m13)*s;
			w=(mat.m32-mat.m23)*s;
			return;
		}
	}
}

void Cquat::normalize(void)
{
	float norm=(float)sqrt(x*x+y*y+z*z+w*w);
	_ASSERT(norm!=0.0F);
	norm=1.0F/norm;
	x*=norm;
	y*=norm;
	z*=norm;
	w*=norm;
}

Cquat interpolate(Cquat const &a,Cquat const &b,float time)
{
	// GRAPHICS GEMS III Tanks. It works.

	float beta;			/* complementary interp parameter */
	float theta;		/* angle between A and B */
	float sin_t, cos_t;	/* sine, cosine of theta */
	float phi;			/* theta plus spins */
	int bflip;			/* use negation of B? */
	float alpha=time;

	/* cosine theta = dot product of A and B */
	cos_t = a.x*b.x + a.y*b.y + a.z*b.z + a.w*b.w;
	/* if B is on opposite hemisphere from A, use -B instead */
 	if (cos_t < 0.0) {
		cos_t = -cos_t;
		bflip = TRUE;
	} else
		bflip = FALSE;
	/* if B is (within precision limits) the same as A,
	 * just linear interpolate between A and B.
	 * Can't do spins, since we don't know what direction to spin.
	 */
	if (1.0F-cos_t<0.0001F)
	{
		beta=1.0F-alpha;
 	} 
	else
	{				
		/* normal case */
 		theta=(float)acos(cos_t);
 		phi=theta ;
 		sin_t=(float)sin(theta);
 		beta=(float)sin(theta - alpha*phi)/sin_t;
 		alpha=(float)sin(alpha*phi)/sin_t;
 	}
	if (bflip)
		alpha = -alpha;
	/* interpolate */
	Cquat res;
 	res.x = beta*a.x + alpha*b.x;
 	res.y = beta*a.y + alpha*b.y;
 	res.z = beta*a.z + alpha*b.z;
 	res.w = beta*a.w + alpha*b.w;
	return res;
}


Cquat interpolate(Cquat const &a,Cquat const &b,Cquat const &c,float ra,float rb)
{
	// TODO : ca marche ptet pas....
	Cquat q1,q2;
	q1=interpolate(a,b,ra*1.0F/(ra+rb));
	q2=interpolate(q1,c,ra+rb);
	return q2;
}

int Cquat::fread(G3FILE *handle)
{
	UW version;
	g3fread(&version,sizeof(version),1,handle);
	switch(version)
	{
	case 1:
		{
		int count=0;
		count+=g3fread(&x,sizeof(x),1,handle);
		count+=g3fread(&y,sizeof(y),1,handle);
		count+=g3fread(&z,sizeof(z),1,handle);
		count+=g3fread(&w,sizeof(w),1,handle);
		if(count==4)
			return(1);
		}
		break;
	default:
		::MessageBox(NULL,"QUAT : UNKNOWN VERSION","ERROR",MB_ICONWARNING|MB_OK);
		exit(0);
		break;

	}
	return(0);
}

int Cquat::fwrite(G3FILE *handle)
{
	UW version=1;
	g3fwrite(&version,sizeof(version),1,handle);
	int count=0;
	count+=g3fwrite(&x,sizeof(x),1,handle);
	count+=g3fwrite(&y,sizeof(y),1,handle);
	count+=g3fwrite(&z,sizeof(z),1,handle);
	count+=g3fwrite(&w,sizeof(w),1,handle);
	if(count==4)
		return(1);
	return(0);
}

